from ._internal.frameworks.paddle import load
from ._internal.frameworks.paddle import save
from ._internal.frameworks.paddle import load_runner
from ._internal.frameworks.paddle import import_from_paddlehub

__all__ = ["import_from_paddlehub", "load", "load_runner", "save"]



import logging
from contextlib import contextmanager

from ..tracing import Tracer

logger = logging.getLogger(__name__)


class NoopTracer(Tracer):
    def __init__(self):
        logger.debug("Tracing is disabled. Initializing no-op tracer")

    @contextmanager
    def span(self, *args, **kwargs):  # pylint: disable=unused-argument
        yield
        return

    @contextmanager
    def async_span(self, *args, **kwargs):  # pylint: disable=unused-argument
        yield
        return



import io
import sys
import typing as t
from typing import TYPE_CHECKING
from urllib.parse import quote

from starlette.requests import Request
from multipart.multipart import parse_options_header
from starlette.responses import Response

from .base import ImageType
from .base import IODescriptor
from ..utils import LazyLoader
from ...exceptions import BadInput
from ...exceptions import InvalidArgument
from ...exceptions import InternalServerError

if TYPE_CHECKING:
    import PIL
    import numpy as np
    import PIL.Image
else:
    np = LazyLoader("np", globals(), "numpy")

    # NOTE: pillow-simd only benefits users who want to do preprocessing
    # TODO: add options for users to choose between simd and native mode
    _exc = f"""\
    `Pillow` is required to use {__name__}
    Instructions: `pip install -U Pillow`
    """
    PIL = LazyLoader("PIL", globals(), "PIL", exc_msg=_exc)
    PIL.Image = LazyLoader("PIL.Image", globals(), "PIL.Image", exc_msg=_exc)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


DEFAULT_PIL_MODE = "RGB"

_Mode = Literal[
    "1", "CMYK", "F", "HSV", "I", "L", "LAB", "P", "RGB", "RGBA", "RGBX", "YCbCr"
]


class Image(IODescriptor[ImageType]):
    """
    `Image` defines API specification for the inputs/outputs of a Service, where either
     inputs will be converted to or outputs will be converted from images as specified
     in your API function signature.

    .. Toy implementation of a transformers service for object detection::
        #obj_detc.py
        import bentoml
        from bentoml.io import Image, JSON
        import bentoml.transformers

        tag='google_vit_large_patch16_224:latest'
        runner = bentoml.transformers.load_runner(tag, tasks='image-classification',
                                                  device=-1,
                                                  feature_extractor="google/vit-large-patch16-224")

        svc = bentoml.Service("vit-object-detection", runners=[runner])

        @svc.api(input=Image(), output=JSON())
        def predict(input_img):
            res = runner.run_batch(input_img)
            return res

    Users then can then serve this service with `bentoml serve`::
        % bentoml serve ./obj_detc.py:svc --auto-reload

        (Press CTRL+C to quit)
        [INFO] Starting BentoML API server in development mode with auto-reload enabled
        [INFO] Serving BentoML Service "vit-object-detection" defined in "obj_detc.py"
        [INFO] API Server running on http://0.0.0.0:5000

    Users can then send a cURL requests like shown in different terminal session::
        # we will run on our input image test.png
        # image can get from http://images.cocodataset.org/val2017/000000039769.jpg
        % curl -H "Content-Type: multipart/form-data" -F
          'fileobj=@test.jpg;type=image/jpeg' http://0.0.0.0:8000/predict

        [{"score":0.8610631227493286,"label":"Egyptian cat"},
         {"score":0.08770329505205154,"label":"tabby, tabby cat"},
         {"score":0.03540956228971481,"label":"tiger cat"},
         {"score":0.004140055272728205,"label":"lynx, catamount"},
         {"score":0.0009498853469267488,"label":"Siamese cat, Siamese"}]%

    Args:
        pilmode (`str`, `optional`, default to `RGB`):
            Color mode for PIL.
        mime_type (`str`, `optional`, default to `image/jpeg`):
            Return MIME type for `starlette.response.Response`

    Returns:
        IO Descriptor that represents either a `np.ndarray` or a `PIL.Image.Image`.
    """

    MIME_EXT_MAPPING: t.Dict[str, str] = {}

    def __init__(
        self,
        pilmode: t.Optional["_Mode"] = DEFAULT_PIL_MODE,
        mime_type: str = "image/jpeg",
    ):
        PIL.Image.init()
        self.MIME_EXT_MAPPING.update({v: k for k, v in PIL.Image.MIME.items()})

        if mime_type.lower() not in self.MIME_EXT_MAPPING:  # pragma: no cover
            raise InvalidArgument(
                f"Invalid Image mime_type '{mime_type}', "
                f"Supported mime types are {', '.join(PIL.Image.MIME.values())} "
            )
        if pilmode is not None and pilmode not in PIL.Image.MODES:  # pragma: no cover
            raise InvalidArgument(
                f"Invalid Image pilmode '{pilmode}', "
                f"Supported PIL modes are {', '.join(PIL.Image.MODES)} "
            )

        self._mime_type = mime_type.lower()
        self._pilmode: t.Optional[_Mode] = pilmode
        self._format = self.MIME_EXT_MAPPING[mime_type]

    def openapi_schema_type(self) -> t.Dict[str, str]:
        return {"type": "string", "format": "binary"}

    def openapi_request_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for incoming requests"""
        return {self._mime_type: {"schema": self.openapi_schema_type()}}

    def openapi_responses_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for outcoming responses"""
        return {self._mime_type: {"schema": self.openapi_schema_type()}}

    async def from_http_request(self, request: Request) -> ImageType:
        content_type, _ = parse_options_header(request.headers["content-type"])
        mime_type = content_type.decode().lower()
        if mime_type == "multipart/form-data":
            form = await request.form()
            bytes_ = await next(iter(form.values())).read()
        elif mime_type.startswith("image/") or content_type == self._mime_type:
            bytes_ = await request.body()
        else:
            raise BadInput(
                f"{self.__class__.__name__} should get `multipart/form-data`, "
                f"`{self._mime_type}` or `image/*`, got {content_type} instead"
            )
        return PIL.Image.open(io.BytesIO(bytes_))

    async def to_http_response(self, obj: ImageType) -> Response:
        if isinstance(obj, np.ndarray):
            image = PIL.Image.fromarray(obj, mode=self._pilmode)
        elif isinstance(obj, PIL.Image.Image):  # type: ignore
            image = obj
        else:
            raise InternalServerError(
                f"Unsupported Image type received: {type(obj)}, `{self.__class__.__name__}`"
                " only supports `np.ndarray` and `PIL.Image`"
            )
        filename = f"output.{self._format.lower()}"

        ret = io.BytesIO()
        image.save(ret, format=self._format)

        # rfc2183
        content_disposition_filename = quote(filename)
        if content_disposition_filename != filename:
            content_disposition = "attachment; filename*=utf-8''{}".format(
                content_disposition_filename
            )
        else:
            content_disposition = f'attachment; filename="{filename}"'
        headers = {"content-disposition": content_disposition}

        return Response(ret.getvalue(), media_type=self._mime_type, headers=headers)



import os
import re
import json
import typing as t
import logging
import tempfile
import functools
import importlib.util
from typing import TYPE_CHECKING
from pathlib import Path
from importlib import import_module
from contextlib import contextmanager

import requests
from filelock import FileLock
from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner
from bentoml.exceptions import NotFound
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import Model
from ..models import JSON_EXT
from ..utils.pkg import get_pkg_version
from ..configuration.containers import BentoMLContainer

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from transformers import PreTrainedModel
    from transformers import PretrainedConfig
    from transformers import TFPreTrainedModel
    from transformers import FlaxPreTrainedModel
    from transformers import PreTrainedTokenizer
    from transformers import PreTrainedTokenizerFast
    from transformers.feature_extraction_utils import PreTrainedFeatureExtractor
    from transformers.models.auto.auto_factory import _BaseAutoModelClass

    from ..models import ModelStore

try:
    import transformers
    from transformers import Pipeline
    from transformers import AutoConfig
    from transformers import AutoTokenizer
    from transformers.file_utils import http_get
    from transformers.file_utils import CONFIG_NAME
    from transformers.file_utils import WEIGHTS_NAME
    from transformers.file_utils import hf_bucket_url
    from transformers.file_utils import http_user_agent
    from transformers.file_utils import TF2_WEIGHTS_NAME
    from transformers.file_utils import FLAX_WEIGHTS_NAME
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """\
        transformers is required in order to use module `bentoml.transformers`.
        Instruction: Install transformers with `pip install transformers`.
        """
    )

_transformers_version = get_pkg_version("transformers")

try:
    from huggingface_hub import HfFolder
except ImportError:
    HfFolder = None

_hfhub_exc = """\
`huggingface_hub` is required to use `bentoml.transformers.import_from_huggingface_hub()`.
Instruction: `pip install huggingface_hub`
"""

_T = t.TypeVar("_T")
_F = t.TypeVar("_F", bound=t.Callable[..., t.Any])
_T_co = t.TypeVar("_T_co", covariant=True)

_ModelType = t.Union["PreTrainedModel", "TFPreTrainedModel", "FlaxPreTrainedModel"]
_TokenizerType = t.Union["PreTrainedTokenizer", "PreTrainedTokenizerFast"]

_FRAMEWORK_ALIASES: t.Dict[str, str] = {"pt": "pytorch", "tf": "tensorflow"}

_AUTOMODEL_PREFIX_MAPPING: t.Dict[str, str] = {
    "pytorch": "AutoModel",
    "tensorflow": "TFAutoModel",
    "flax": "FlaxAutoModel",
}

_AUTOMODEL_LM_HEAD_MAPPING: t.Dict[str, str] = {
    "causal": "ForCausalLM",
    "masked": "ForMaskedLM",
    "seq2seq": "ForSeq2SeqLM",
    "sequence-classification": "ForSequenceClassification",
    "question-answering": "ForQuestionAnswering",
    "token-classification": "ForTokenClassification",
    "multiple-choice": "ForMultipleChoice",
    "next-sentence-prediction": "ForNextSentencePrediction",
    "image-classification": "ForImageClassification",
    "audio-classification": "ForAudioClassification",
    "ctc": "ForCTC",
    "speech-seq2seq": "ForSpeechSeq2Seq",
    "object-detection": "ForObjectDetection",
}


class _GeneratorContextManager(t.ContextManager[_T_co]):
    def __call__(self, func: _F) -> _F:
        ...


def _load_autoclass(framework: str, lm_head: str) -> "_BaseAutoModelClass":
    if (
        framework not in _AUTOMODEL_PREFIX_MAPPING
        and framework not in _FRAMEWORK_ALIASES
    ):
        raise AttributeError(
            f"{framework} is either invalid aliases "
            "or not supported by transformers. "
            "Accepted: pt(alias to pytorch), "
            "tf(alias to tensorflow), and flax."
        )
    if lm_head not in _AUTOMODEL_LM_HEAD_MAPPING:
        raise AttributeError(
            f"`{lm_head}` alias for lm_head is invalid."
            f" Accepted: {[*_AUTOMODEL_LM_HEAD_MAPPING.keys()]}."
            " If you need any other AutoModel type provided by transformers,"
            " feel free to open a PR at https://github.com/bentoml/BentoML."
        )
    framework_prefix = (
        _FRAMEWORK_ALIASES[framework] if framework in _FRAMEWORK_ALIASES else framework
    )
    class_inst = f"{_AUTOMODEL_PREFIX_MAPPING[framework_prefix]}{_AUTOMODEL_LM_HEAD_MAPPING[lm_head]}"  # noqa
    try:
        return getattr(import_module("transformers"), class_inst)
    except AttributeError as e:
        raise BentoMLException(
            f"{e}\n\nPlease refers "
            f"to https://huggingface.co/transformers/model_doc/auto.html."
        )


def _clean_name(name: str) -> str:
    return re.sub(r"\W|^(?=\d)-", "_", name)


def _check_flax_supported() -> None:  # pragma: no cover
    _supported: bool = _transformers_version.startswith("4")
    _flax_available = (
        importlib.util.find_spec("jax") is not None
        and importlib.util.find_spec("flax") is not None
    )
    if not _supported:
        logger.warning(
            "Detected transformers version: "
            f"{_transformers_version}, which "
            "doesn't have supports for Flax. "
            "Update `transformers` to 4.x and "
            "above to have Flax supported."
        )
    else:
        if _flax_available:
            _jax_version = get_pkg_version("jax")
            _flax_version = get_pkg_version("flax")
            logger.info(
                f"JAX version {_jax_version}, "
                f"Flax version {_flax_version} available."
            )
        else:
            logger.warning(
                "No versions of Flax or Jax are found under "
                "the current machine. In order to use "
                "Flax with transformers 4.x and above, "
                "refers to https://github.com/google/flax#quick-install"
            )


_SAVE_CONFLICTS_ERR = """\
When `tokenizer={tokenizer}`, model should be of type
Union[`PreTrainedModel`, `TFPreTrainedModel`, `FlaxPreTrainedModel`]. Currently
`type(model)={model}`

If you want to save the weight directly from
`transformers` and save it to BentoML do:
    `bentoml.transformers.import_from_huggingface_hub('bert-base-uncased')`.

If you are training a model from scratch using transformers, to save into BentoML do:
    `bentoml.transformers.save('bert_model', model=my_bert_model, tokenizer=my_tokenizer)`

If you want to import directly from a `transformers.pipeline` then do:
    # pipeline = transformers.pipelines('sentiment-analysis')
    `bentoml.transoformers.save("senta-pipe", pipeline)
"""


@inject
def load(
    tag: t.Union[str, Tag],
    from_tf: bool = False,
    from_flax: bool = False,
    framework: str = "pt",
    lm_head: str = "causal",
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    **kwargs: str,
) -> t.Tuple[
    "PretrainedConfig",
    t.Union["PreTrainedModel", "TFPreTrainedModel", "FlaxPreTrainedModel"],
    t.Optional[t.Union["PreTrainedTokenizer", "PreTrainedTokenizerFast"]],
]:
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.
        from_tf (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Load the model weights from a TensorFlow checkpoint save file
        from_flax (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Load the model weights from a Flax checkpoint save file
        framework (`str`, default to `pt`):
            Given frameworks supported by transformers: PyTorch, Tensorflow, Flax
        lm_head (`str`, default to `causal`):
            Language model head for your model. For most usecase causal are applied.
             Refers to huggingface.co/transformers for more details on which type of
             language model head is applied to your use case and model.
        kwargs (:obj:`str`, `Optional`):
            kwargs that can be parsed to transformers Models instance.

    Returns:
        a Tuple containing `model` and `tokenizer` for your given model saved at BentoML
         modelstore.

    Examples::
        import bentoml.transformers
        model, tokenizer = bentoml.transformers.load('custom_gpt2', framework="flax",
                                                     lm_head="masked")
    """
    _check_flax_supported()  # pragma: no cover
    model = model_store.get(tag)
    _model, _tokenizer = model.info.options["model"], model.info.options["tokenizer"]

    if _tokenizer != "na":
        tokenizer = getattr(import_module("transformers"), _tokenizer).from_pretrained(
            model.path, from_tf=from_tf, from_flax=from_flax
        )
    else:
        tokenizer = None
    config = AutoConfig.from_pretrained(model.path)

    try:
        # Cover cases where some model repo doesn't include a model
        #  name under their config.json. An example is
        #  google/bert_uncased-L-2-H-128-A-2
        t_model = _load_autoclass(framework, lm_head).from_pretrained(
            model.path,
            config=config,
            **kwargs,
        )
    except (AttributeError, BentoMLException):  # noqa
        t_model = getattr(import_module("transformers"), _model).from_pretrained(
            model.path,
            config=config,
            from_tf=from_tf,
            from_flax=from_flax,
            **kwargs,
        )
    return config, t_model, tokenizer


def _download_from_hub(
    hf_url: str,
    output_dir: str,
    force_download: bool = False,
    proxies: t.Union[str, None] = None,
    etag_timeout: int = 10,
    resume_download: bool = False,
    user_agent: t.Union[t.Dict[str, str], str, None] = None,
    use_auth_token: t.Union[bool, str, None] = None,
) -> None:
    """
    Modification of https://github.com/huggingface/transformers/blob/master/src/transformers/file_utils.py
    """
    if HfFolder is None:
        raise BentoMLException(_hfhub_exc)
    headers = {"user-agent": http_user_agent(user_agent)}
    if isinstance(use_auth_token, str):
        headers["authorization"] = f"Bearer {use_auth_token}"
    elif use_auth_token:
        token = HfFolder.get_token()
        if token is None:
            raise EnvironmentError(
                "You specified use_auth_token=True, "
                "but a huggingface token was not found."
            )
        headers["authorization"] = f"Bearer {token}"

    url_to_download = hf_url
    etag = None
    try:
        r = requests.head(
            hf_url,
            headers=headers,
            allow_redirects=False,
            proxies=proxies,
            timeout=etag_timeout,
        )
        r.raise_for_status()
        etag = r.headers.get("X-Linked-Etag") or r.headers.get("ETag")
        # We favor a custom header indicating the etag of the linked resource, and
        # we fallback to the regular etag header.
        # If we don't have any of those, raise an error.
        if etag is None:
            raise OSError(
                "Distant resource does not have an ETag, "
                "we won't be able to reliably ensure reproducibility."
            )
        # In case of a redirect,
        # save an extra redirect on the request.get call,
        # and ensure we download the exact atomic version even if it changed
        # between the HEAD and the GET (unlikely, but hey).
        if 300 <= r.status_code <= 399:
            url_to_download = r.headers["Location"]
    except (requests.exceptions.SSLError, requests.exceptions.ProxyError):
        # Actually raise for those subclasses of ConnectionError
        raise
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        # Otherwise, our Internet connection is down.
        # etag is None
        pass
    fname = Path(output_dir, hf_url.split("/")[-1])
    logger.info(f"Downloading {hf_url} to {str(fname)}")
    fpath = Path(output_dir, fname)
    # From now on, etag is not None.
    if fpath.exists() and not force_download:
        return

    # Prevent parallel downloads of the same file with a lock.
    lock_path = fpath.with_suffix(".lock")
    with FileLock(str(lock_path)):

        # If the download just completed while the lock was activated.
        if os.path.exists(fpath) and not force_download:
            # Even if returning early like here, the lock will be released.
            return

        _tmp_file_manager: t.Union[
            t.Callable[[], _GeneratorContextManager[t.BinaryIO]],
            t.Callable[[t.Any, t.Any], t.IO[str]],
        ]

        if resume_download:
            incomplete_path = fpath.with_suffix(".incomplete")

            @contextmanager
            def _resume_file_manager() -> t.Generator[t.BinaryIO, None, None]:
                with open(incomplete_path, "ab") as f:
                    yield f

            _tmp_file_manager = _resume_file_manager
            if os.path.exists(incomplete_path):
                resume_size = os.stat(incomplete_path).st_size
            else:
                resume_size = 0
        else:
            _tmp_file_manager = functools.partial(
                tempfile.NamedTemporaryFile, mode="wb", dir=output_dir, delete=False
            )
            resume_size = 0
    with _tmp_file_manager() as temp_file:
        http_get(
            url_to_download,
            temp_file,
            proxies=proxies,
            resume_size=resume_size,
            headers=headers,
        )

    logger.info(f"storing {hf_url} at {output_dir}")
    os.replace(temp_file.name, str(fname))

    # NamedTemporaryFile creates a file with
    #  hardwired 0600 perms (ignoring umask), so fixing it.
    umask = os.umask(0o666)
    os.umask(umask)
    os.chmod(fname, 0o666 & ~umask)
    logger.info(f"creating metadata file for {fpath.name}")
    meta = {"url": hf_url, "etag": etag}
    meta_path = fpath.with_suffix(f".metadata{JSON_EXT}")
    with meta_path.open("w") as meta_file:
        json.dump(meta, meta_file)


def _save(
    name: str,
    *,
    model_identifier: t.Union[str, _ModelType, "Pipeline"],
    tokenizer: t.Optional[_TokenizerType],
    metadata: t.Optional[t.Dict[str, t.Any]],
    keep_download_from_hub: bool,
    model_store: "ModelStore",
    **transformers_options_kwargs: str,
) -> Tag:
    _check_flax_supported()  # pragma: no cover
    context: t.Dict[str, t.Any] = {
        "framework_name": "transformers",
        "pip_dependencies": [f"transformers=={_transformers_version}"],
    }

    if isinstance(model_identifier, str):
        try:
            info = model_store.get(name)
            if not keep_download_from_hub:
                logger.warning(
                    f"{name} is found under BentoML modelstore.\nFor most usecases of using pretrained model,"
                    f" you don't have to redownload the model. returning {info.tag}...\nIf you still insist on downloading,"
                    " then specify `keep_download_from_hub=True` in `import_from_huggingface_hub`"
                )
                return info.tag
            else:
                pass
        except (NotFound, FileNotFoundError):
            pass

    _model = Model.create(
        name,
        module=__name__,
        context=context,
        options=None,
        metadata=metadata,
    )

    if tokenizer is not None:
        assert not isinstance(model_identifier, str) or isinstance(
            model_identifier, Pipeline
        ), _SAVE_CONFLICTS_ERR.format(tokenizer=tokenizer, model=type(model_identifier))
        _model.info.options = {
            "model": model_identifier.__class__.__name__,
            "tokenizer": tokenizer.__class__.__name__,
            "pipeline": False,
        }
        model_identifier.save_pretrained(_model.path)
        tokenizer.save_pretrained(_model.path)
    else:
        if isinstance(model_identifier, Pipeline):
            # model_identifier is a Pipeline
            _model.info.options = {
                "model": model_identifier.model.__class__.__name__,
                "tokenizer": model_identifier.tokenizer.__class__.__name__,
                "pipeline": True,
            }
            model_identifier.save_pretrained(_model.path)
        else:
            from_tf = transformers_options_kwargs.pop("from_tf", False)
            from_flax = transformers_options_kwargs.pop("from_flax", False)
            revision = transformers_options_kwargs.pop("revision", None)
            mirror = transformers_options_kwargs.pop("mirror", None)
            proxies = transformers_options_kwargs.pop("proxies", None)
            use_auth_token = transformers_options_kwargs.pop("use_auth_token", None)
            force_download = bool(
                transformers_options_kwargs.pop("force_download", False)
            )
            resume_download = bool(
                transformers_options_kwargs.pop("resume_download", True)
            )

            # download config file
            config_file_url = hf_bucket_url(
                model_identifier,
                filename=CONFIG_NAME,
                revision=revision,
                mirror=mirror,
            )
            _download_from_hub(
                config_file_url,
                output_dir=_model.path,
                proxies=proxies,
                use_auth_token=use_auth_token,
                force_download=force_download,
                resume_download=resume_download,
            )

            # download weight file set correct filename
            if from_tf:
                weight_filename = TF2_WEIGHTS_NAME
            elif from_flax:
                weight_filename = FLAX_WEIGHTS_NAME
            else:
                weight_filename = WEIGHTS_NAME

            weight_file_url = hf_bucket_url(
                model_identifier,
                filename=weight_filename,
                revision=revision,
                mirror=mirror,
            )
            _download_from_hub(
                weight_file_url,
                output_dir=_model.path,
                proxies=proxies,
                use_auth_token=use_auth_token,
                force_download=force_download,
                resume_download=resume_download,
            )

            # _model.info.options will be None here
            with open(_model.path_of(CONFIG_NAME), "r", encoding="utf-8") as of:
                _config = json.loads(of.read())
                try:
                    arch = "".join(_config["architectures"])
                    if from_tf:
                        _model.info.options["model"] = f"TF{arch}"
                    elif from_flax:
                        _model.info.options["model"] = f"Flax{arch}"
                    else:
                        _model.info.options["model"] = arch
                except KeyError:
                    _model.info.options["model"] = ""
            _model.info.options["pipeline"] = False

            # NOTE: With Tokenizer there are way too many files
            #  to be included, per frameworks. Thus we will load
            #  a Tokenizer instance, then save it to path.
            try:
                _tokenizer_inst = AutoTokenizer.from_pretrained(model_identifier)
                _tokenizer_inst.save_pretrained(_model.path)
                _model.info.options["tokenizer"] = type(_tokenizer_inst).__name__
            except ValueError:
                _model.info.options["tokenizer"] = "na"

    _model.save(model_store)
    return _model.tag


@inject
def save(
    name: str,
    *,
    model: _ModelType,
    tokenizer: t.Optional[_TokenizerType] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    **transformers_options_kwargs: str,
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`t.Union["PreTrainedModel", "TFPreTrainedModel", "FlaxPreTrainedModel"]`,
               required):
            Model instance provided by transformers. This can be retrieved from their
             `AutoModel` class. You can also use any type of models/automodel provided
             by transformers. Refers to https://huggingface.co/transformers/main_classes/model.html
        tokenizer (`t.Union["PreTrainedTokenizer", "PreTrainedTokenizerFast"]`):
            Tokenizer instance provided by transformers. This can be retrieved from
             their `AutoTokenizer` class. You can also use any type of Tokenizer
             accordingly to your usecase provided by transformers. Refers to
             https://huggingface.co/transformers/main_classes/tokenizer.html
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.
        from_tf (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Load the model weights from a TensorFlow checkpoint save file
        from_flax (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Load the model weights from a Flax checkpoint save file
        revision(:obj:`str`, `Optional`, defaults to :obj:`"main"`):
            The specific model version to use. It can be a branch name, a tag name, or a
            commit id, since we use a git-based system for storing models and other
            artifacts on huggingface.co, so ``revision`` can be any
            identifier allowed by git.
        mirror(:obj:`str`, `Optional`):
            Mirror source to accelerate downloads in China. If you are from China and
            have an accessibility problem, you can set this option to resolve it. Note
            that we do not guarantee the timeliness or safety. Please refer to the
            mirror site for more information.
        proxies (:obj:`Dict[str, str], `Optional`):
            A dictionary of proxy servers to use by protocol or endpoint, e.g.
            :obj:`{'http': 'foo.bar:3128', 'http://hostname': 'foo.bar:4012'}`. The
            proxies are used on each request.
        use_auth_token (:obj:`str` or `bool`, `Optional`):
            The token to use as HTTP bearer authorization for remote files. If
            :obj:`True`, will use the token generated when running
            :obj:`transformers-cli login` (stored in :obj:`~/.huggingface`).
        force_download (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Whether or not to force the (re-)download of the model weights and
            configuration files, overriding the cached versions if they exist.
        resume_download (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Whether or not to delete incompletely received files. Will attempt to resume
            the download if such a file exists.

        .. note::
            some parameters are direct port from `from_pretrained()` arguments. This
            ensures that when doing save operations we don't actually load the model
            class, which can take a while to do so.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
        from transformers import AutoModelForQuestionAnswering, AutoTokenizer
        import bentoml.transformers

        model = AutoModelForQuestionAnswering.from_pretrained("gpt2", from_flax=True)
        tokenizer = AutoTokenizer.from_pretrained("gpt2", from_flax=True)
        # custom training and modification goes here

        tag = bentoml.transformers.save("flax_gpt2", model=model, tokenizer=tokenizer)
    """
    if isinstance(model, str) and not tokenizer:
        raise EnvironmentError(
            "If you want to import model directly from huggingface hub"
            " please use `import_from_huggingface_hub` instead. `save` should"
            " only be used for saving custom pretrained model and tokenizer"
        )
    return _save(
        name=name,
        model_identifier=model,
        tokenizer=tokenizer,
        metadata=metadata,
        keep_download_from_hub=False,
        model_store=model_store,
        **transformers_options_kwargs,
    )


@inject
def import_from_huggingface_hub(
    name: str,
    *,
    save_namespace: t.Optional[str] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    keep_download_from_hub: bool = False,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    **transformers_options_kwargs: str,
) -> Tag:
    """
    Import a model from hugging face hub and save it to bentoml modelstore.

    Args:
        name (`str`):
            Model name retrieved from huggingface hub. This shouldn't be a model
             instance. If you would like to save a model instance refers to
             `~bentoml.transformers.save` for more information.
        save_namespace (`str`, default to given `name`):
            Name to save model to BentoML modelstore.
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.
        keep_download_from_hub (`bool`, `optional`, default to `False`):
            Whether to re-download pretrained model from hub.
        from_tf (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Load the model weights from a TensorFlow checkpoint save file
        from_flax (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Load the model weights from a Flax checkpoint save file
        revision(:obj:`str`, `Optional`, defaults to :obj:`"main"`):
            The specific model version to use. It can be a branch name, a tag name, or a
            commit id, since we use a git-based system for storing models and other
            artifacts on huggingface.co, so ``revision`` can be any identifier allowed
            by git.
        mirror(:obj:`str`, `Optional`):
            Mirror source to accelerate downloads in China. If you are from China and
            have an accessibility problem, you can set this option to resolve it. Note
            that we do not guarantee the timeliness or safety. Please refer to the
            mirror site for more information.
        proxies (:obj:`Dict[str, str], `Optional`):
            A dictionary of proxy servers to use by protocol or endpoint, e.g.
            :obj:`{'http': 'foo.bar:3128', 'http://hostname': 'foo.bar:4012'}`. The
            proxies are used on each request.
        use_auth_token (:obj:`str` or `bool`, `Optional`):
            The token to use as HTTP bearer authorization for remote files. If
            :obj:`True`, will use the token generated when running
            :obj:`transformers-cli login` (stored in :obj:`~/.huggingface`).
        force_download (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Whether or not to force the (re-)download of the model weights and
            configuration files, overriding the cached versions if they exist.
        resume_download (:obj:`bool`, `Optional`, defaults to :obj:`False`):
            Whether or not to delete incompletely received files. Will attempt to resume
            the download if such a file exists.

        .. note::
            some parameters are direct port from `from_pretrained()` arguments. This
            ensures that when doing save operations we don't actually load the model
            class, which can take up a lot of time and resources.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
        from transformers import AutoModelForQuestionAnswering, AutoTokenizer
        import bentoml.transformers

        tag = bentoml.transformers.import_from_huggingface_hub("gpt2", from_tf=True)
    """
    save_namespace = _clean_name(name) if save_namespace is None else save_namespace
    return _save(
        name=save_namespace,
        model_identifier=name,
        tokenizer=None,
        keep_download_from_hub=keep_download_from_hub,
        metadata=metadata,
        model_store=model_store,
        **transformers_options_kwargs,
    )


class _TransformersRunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        tasks: str,
        *,
        framework: str,
        lm_head: str,
        device: int,
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
        **pipeline_kwargs: t.Any,
    ):
        in_store_tag = model_store.get(tag).tag
        self._tag = in_store_tag
        super().__init__(str(in_store_tag), resource_quota, batch_options)

        try:
            transformers.pipelines.check_task(tasks)
        except KeyError as e:
            raise BentoMLException(
                f"{e}, as `{tasks}` is not recognized by transformers."
            )
        self._tasks = tasks
        self._model_store = model_store
        self._framework = framework
        self._lm_head = lm_head
        self._device = device
        self._pipeline_kwargs = pipeline_kwargs

        # pipeline arguments
        self._feature_extractor = pipeline_kwargs.pop(
            "feature_extractor", None
        )  # type: t.Optional[t.Union[str, PreTrainedFeatureExtractor]]
        self._revision = pipeline_kwargs.pop("revision", None)  # type: t.Optional[str]
        self._use_fast = pipeline_kwargs.pop("use_fast", True)  # type: bool
        self._use_auth_token = pipeline_kwargs.pop(
            "use_auth_token", None
        )  # type: t.Optional[t.Union[str, bool]]
        self._model_kwargs = pipeline_kwargs.pop(
            "model_kwargs", {}
        )  # type: t.Dict[str, t.Any]
        self._kwargs = pipeline_kwargs

    @property
    def required_models(self) -> t.List[Tag]:
        return [self._tag]

    @property
    def num_concurrency(self) -> int:
        return self.num_replica

    @property
    def num_replica(self) -> int:
        # TODO: supports multiple GPUS
        return 1

    # pylint: disable=arguments-differ,attribute-defined-outside-init
    def _setup(self) -> None:  # type: ignore[override]
        try:
            _ = self._model_store.get(self._tag)
            self._config, self._model, self._tokenizer = load(
                self._tag,
                model_store=self._model_store,
                from_flax=False,
                from_tf="tf" in self._framework,
                framework=self._framework,
                lm_head=self._lm_head,
            )
        except FileNotFoundError:
            self._config, self._model, self._tokenizer = None, None, None
        self._pipeline: "Pipeline" = transformers.pipeline(
            self._tasks,
            config=self._config,
            model=self._model,
            tokenizer=self._tokenizer,
            framework=self._framework,
            feature_extractor=self._feature_extractor,
            revision=self._revision,
            use_fast=self._use_fast,
            use_auth_token=self._use_auth_token,
            model_kwargs=self._model_kwargs,
            device=self._device,
            **self._kwargs,
        )

    # pylint: disable=arguments-differ
    def _run_batch(self, input_data: t.Union[_T, t.List[_T]]) -> t.Union[_T, t.List[_T]]:  # type: ignore[override]  # noqa
        res = self._pipeline(input_data)  # type: t.Union[_T, t.List[_T]]
        return res


def load_runner(
    tag: t.Union[str, Tag],
    *,
    tasks: str,
    framework: str = "pt",
    lm_head: str = "casual",
    device: int = -1,
    resource_quota: t.Optional[t.Dict[str, t.Any]] = None,
    batch_options: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    **pipeline_kwargs: t.Any,
) -> "_TransformersRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.transformers.load_runner` implements a Runner class
    that wrap around a transformers pipeline, which optimize it for the BentoML runtime.

    .. warning::
        `load_runner` will try to load the model from given `tag`. If the model does not
         exists, then BentoML will fallback to initialize pipelines from transformers,
         thus files will be loaded from huggingface cache.


    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        tasks (`str`):
            Given tasks for pipeline. Refers to https://huggingface.co/transformers/task_summary.html
             for more information.
        framework (`str`, default to `pt`):
            Given frameworks supported by transformers: PyTorch, Tensorflow
        device (`int`, `optional`, default to `-1`):
            Default GPU devices to be used by runner.
        lm_head (`str`, default to `causal`):
            Language model head for your model. For most usecase causal are applied.
             Refers to huggingface.co/transformers for more details on which type of
             language model head is applied to your usecase.
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for `bentoml.transformers` model

    Examples::
        import transformers
        import bentoml.transformers

        runner = bentoml.transformers.load_runner("gpt2:latest", tasks='zero-shot-classification',
                                                  framework=tf)
        runner.run_batch(["In today news, ...", "The stocks market seems ..."])
    """
    return _TransformersRunner(
        tag=tag,
        tasks=tasks,
        framework=framework,
        lm_head=lm_head,
        device=device,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
        **pipeline_kwargs,
    )



# type: ignore[reportMissingTypeStubs]
import typing as t
import functools
from typing import TYPE_CHECKING
from pathlib import Path

import numpy as np
import cloudpickle
from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml.exceptions import MissingDependencyException

from ..models import Model
from ..models import H5_EXT
from ..models import PKL_EXT
from ..models import HDF5_EXT
from ..models import JSON_EXT
from ..models import SAVE_NAMESPACE
from ..utils.tensorflow import get_tf_version
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:  # pragma: no cover
    from tensorflow.python.framework.ops import Graph
    from tensorflow.python.client.session import BaseSession

    from ..models import ModelStore

try:
    import tensorflow as tf

    # TODO: separation of Keras and Tensorflow
    # https://twitter.com/fchollet/status/1404967230048149506?lang=en
    import tensorflow.keras as keras
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """\
        `tensorflow` is required to use `bentoml.keras`, since
        we will use Tensorflow as Keras backend.\n
        Instruction: Refers to https://www.tensorflow.org/install for
        more information of your use case.
        """
    )

from .tensorflow import _TensorflowRunner  # type: ignore[reportPrivateUsage]

_tf_version = get_tf_version()
TF2 = _tf_version.startswith("2")


# Global instance of tf.Session()
_graph: "Graph" = tf.compat.v1.get_default_graph()
_sess: "BaseSession" = tf.compat.v1.Session(graph=_graph)


_CUSTOM_OBJ_FNAME = f"{SAVE_NAMESPACE}_custom_objects{PKL_EXT}"
_SAVED_MODEL_FNAME = f"{SAVE_NAMESPACE}{H5_EXT}"
_MODEL_WEIGHT_FNAME = f"{SAVE_NAMESPACE}_weights{HDF5_EXT}"
_MODEL_JSON_FNAME = f"{SAVE_NAMESPACE}_json{JSON_EXT}"


def get_session() -> "BaseSession":
    """
    Return TF1 sessions for bentoml.keras. Users should
     use this for TF1 in order to run prediction.
    Example::
        session = bentoml.keras.get_session()
        # Initialize variables in the graph/model
        session.run(tf.global_variables_initializer())
        with session.as_default():
            loaded = bentoml.keras.load(tag, model_store=modelstore)
    """  # noqa: LN001
    return _sess


@inject
def load(
    tag: t.Union[str, Tag],
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "keras.Model":
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of `xgboost.core.Booster` from BentoML modelstore.

    Examples::
    """  # noqa

    model = model_store.get(tag)
    default_custom_objects = None
    if model.info.options["custom_objects"]:
        assert Path(model.path_of(_CUSTOM_OBJ_FNAME)).is_file()
        with Path(model.path_of(_CUSTOM_OBJ_FNAME)).open("rb") as dcof:
            default_custom_objects = cloudpickle.load(dcof)

    with get_session().as_default():
        if model.info.options["store_as_json"]:
            assert Path(model.path_of(_MODEL_JSON_FNAME)).is_file()
            with Path(model.path_of(_MODEL_JSON_FNAME)).open("r") as jsonf:
                model_json = jsonf.read()
            model = keras.models.model_from_json(
                model_json, custom_objects=default_custom_objects
            )
        else:
            model = keras.models.load_model(
                model.path_of(_SAVED_MODEL_FNAME),
                custom_objects=default_custom_objects,
            )
        try:
            # if model is a dictionary
            return model["model"]
        except TypeError:
            return model


@inject
def save(
    name: str,
    model: "keras.Model",
    *,
    store_as_json: t.Optional[bool] = False,
    custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`tensorflow.keras.Model`):
            Instance of model to be saved
        store_as_json (`bool`, `optional`, default to `False`):
            Whether to store Keras model as JSON and weights
        custom_objects (`GenericDictType`, `optional`, default to `None`):
            Dictionary of Keras custom objects for mod
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
    """  # noqa
    tf.compat.v1.keras.backend.get_session()
    context: t.Dict[str, t.Any] = {
        "framework_name": "keras",
        "pip_dependencies": [f"tensorflow=={_tf_version}"],
    }
    options = {
        "store_as_json": store_as_json,
        "custom_objects": True if custom_objects is not None else False,
    }
    _model = Model.create(
        name,
        module=__name__,
        options=options,
        context=context,
        metadata=metadata,
    )

    if custom_objects is not None:
        with Path(_model.path_of(_CUSTOM_OBJ_FNAME)).open("wb") as cof:
            cloudpickle.dump(custom_objects, cof)
    if store_as_json:
        with Path(_model.path_of(_MODEL_JSON_FNAME)).open("w") as jf:
            jf.write(model.to_json())
        model.save_weights(_model.path_of(_MODEL_WEIGHT_FNAME))
    else:
        model.save(_model.path_of(_SAVED_MODEL_FNAME))

    _model.save(model_store)

    return _model.tag


class _KerasRunner(_TensorflowRunner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        predict_fn_name: str,
        device_id: str,
        predict_kwargs: t.Optional[t.Dict[str, t.Any]],
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        super().__init__(
            tag=tag,
            predict_fn_name=predict_fn_name,
            device_id=device_id,
            partial_kwargs=predict_kwargs,
            resource_quota=resource_quota,
            batch_options=batch_options,
            model_store=model_store,
        )
        self._session = get_session()

    @property
    def session(self) -> "BaseSession":
        return self._session

    # pylint: disable=attribute-defined-outside-init
    def _setup(self) -> None:
        self._session.config = self._config_proto
        self._model = load(self._tag, model_store=self._model_store)
        raw_predict_fn = getattr(self._model, self._predict_fn_name)
        self._predict_fn = functools.partial(raw_predict_fn, **self._partial_kwargs)

    # pylint: disable=arguments-differ
    def _run_batch(  # type: ignore[override]
        self,
        input_data: t.Union[
            t.List[t.Union[int, float]], "np.ndarray[t.Any, np.dtype[t.Any]]", tf.Tensor
        ],
    ) -> t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", tf.Tensor]:
        if not isinstance(input_data, (np.ndarray, tf.Tensor)):
            input_data = np.array(input_data)
        with tf.device(self._device_id):
            if TF2:
                tf.compat.v1.global_variables_initializer()
            else:
                self._session.run(tf.compat.v1.global_variables_initializer())
                with get_session().as_default():
                    self._predict_fn(input_data)
            return self._predict_fn(input_data)


@inject
def load_runner(
    tag: t.Union[str, Tag],
    *,
    predict_fn_name: str = "predict",
    device_id: str = "CPU:0",
    predict_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
    resource_quota: t.Union[None, t.Dict[str, t.Any]] = None,
    batch_options: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "_KerasRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.tensorflow.load_runner` implements a Runner class that
    wrap around a Keras model, with Tensorflow as backend, which optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        predict_fn_name (`str`, default to `predict`):
            inference function to be used.
        predict_kwargs (`t.Dict[str, t.Any]`, `optional`, default to `None`):
            Dictionary of `predict()` kwargs that can be shared across different model.
        device_id (`t.Union[str, int, t.List[t.Union[str, int]]]`, `optional`, default to `CPU:0`):
            Optional devices to put the given model on. Refers to https://www.tensorflow.org/api_docs/python/tf/config
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for `bentoml.keras` model

    Examples::
    """  # noqa: LN001
    return _KerasRunner(
        tag=tag,
        predict_fn_name=predict_fn_name,
        device_id=device_id,
        predict_kwargs=predict_kwargs,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )



import os
import typing as t
import datetime
from abc import ABC
from abc import abstractmethod
from contextlib import contextmanager

import fs
import fs.errors
from fs.base import FS

from .types import Tag
from .types import PathType
from ..exceptions import NotFound
from ..exceptions import BentoMLException

T = t.TypeVar("T")


class StoreItem(ABC):
    @property
    @abstractmethod
    def tag(self) -> Tag:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_fs(cls: t.Type[T], item_fs: FS) -> T:
        raise NotImplementedError

    @property
    @abstractmethod
    def creation_time(self) -> datetime.datetime:
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}(tag="{self.tag}")'


Item = t.TypeVar("Item", bound=StoreItem)


class Store(ABC, t.Generic[Item]):
    """An FsStore manages items under the given base filesystem.

    Note that FsStore has no consistency checks; it assumes that no direct modification
    of the files in its directory has occurred.

    """

    _fs: FS
    _item_type: t.Type[Item]

    @abstractmethod
    def __init__(self, base_path: t.Union[PathType, FS], item_type: t.Type[Item]):
        self._item_type = item_type
        if isinstance(base_path, os.PathLike):
            base_path = base_path.__fspath__()
        self._fs = fs.open_fs(base_path)

    def list(self, tag: t.Optional[t.Union[Tag, str]] = None) -> t.List[Item]:
        if not tag:
            return [
                ver for _d in sorted(self._fs.listdir("/")) for ver in self.list(_d)
            ]

        _tag = Tag.from_taglike(tag)
        if _tag.version is None:
            if not self._fs.isdir(_tag.name):
                raise NotFound(
                    f"no {self._item_type.__name__}s with name '{_tag.name}' found"
                )

            tags = sorted(
                [
                    Tag(_tag.name, f.name)
                    for f in self._fs.scandir(_tag.name)
                    if f.is_dir
                ]
            )
            return [self._get_item(t) for t in tags]
        else:
            return [self._get_item(_tag)] if self._fs.isdir(_tag.path()) else []

    def _get_item(self, tag: Tag) -> Item:
        """
        Creates a new instance of Item that represents the item with tag `tag`.
        """
        return self._item_type.from_fs(self._fs.opendir(tag.path()))

    def get(self, tag: t.Union[Tag, str]) -> Item:
        """
        store.get("my_bento")
        store.get("my_bento:v1.0.0")
        store.get(Tag("my_bento", "latest"))
        """
        _tag = Tag.from_taglike(tag)
        if _tag.version is None or _tag.version == "latest":
            try:
                _tag.version = self._fs.readtext(_tag.latest_path())
            except fs.errors.ResourceNotFound:
                raise NotFound(
                    f"no {self._item_type.__name__}s with name '{_tag.name}' exist in BentoML store {self._fs}"
                )

        path = _tag.path()
        if self._fs.exists(path):
            return self._get_item(_tag)

        matches = self._fs.glob(f"{path}*/")
        counts = matches.count().directories
        if counts == 0:
            raise NotFound(
                f"{self._item_type.__name__} '{tag}' is not found in BentoML store {self._fs}"
            )
        elif counts == 1:
            match = next(iter(matches))
            return self._get_item(Tag(_tag.name, match.info.name))
        else:
            vers: t.List[str] = []
            for match in matches:
                vers += match.info.name
            raise BentoMLException(
                f"multiple versions matched by {_tag.version}: {vers}"
            )

    @contextmanager
    def register(self, tag: t.Union[str, Tag]):
        _tag = Tag.from_taglike(tag)

        item_path = _tag.path()
        if self._fs.exists(item_path):
            raise BentoMLException(
                f"Item '{_tag}' already exists in the store {self._fs}"
            )
        self._fs.makedirs(item_path)
        try:
            yield self._fs.getsyspath(item_path)
        finally:
            # item generation is most likely successful, link latest path
            if (
                not self._fs.exists(_tag.latest_path())
                or self.get(_tag).creation_time > self.get(_tag.name).creation_time
            ):
                with self._fs.open(_tag.latest_path(), "w") as latest_file:
                    latest_file.write(_tag.version)

    def delete(self, tag: t.Union[str, Tag]) -> None:
        _tag = Tag.from_taglike(tag)

        if not self._fs.exists(_tag.path()):
            raise NotFound(f"{self._item_type.__name__} '{tag}' not found")

        self._fs.removetree(_tag.path())
        if self._fs.isdir(_tag.name):
            versions = self.list(_tag.name)
            if len(versions) == 0:
                # if we've removed all versions, remove the directory
                self._fs.removetree(_tag.name)
            else:
                new_latest = sorted(versions, key=lambda x: x.creation_time)[-1]
                # otherwise, update the latest version
                assert new_latest.tag.version is not None
                self._fs.writetext(_tag.latest_path(), new_latest.tag.version)



import json
import typing as t
import asyncio
import logging
from typing import TYPE_CHECKING
from functools import partial

from simple_di import inject
from simple_di import Provide

from bentoml._internal.runner.utils import Params
from bentoml._internal.runner.utils import PAYLOAD_META_HEADER
from bentoml._internal.runner.utils import multipart_to_payload_params
from bentoml._internal.runner.container import AutoContainer
from bentoml._internal.marshal.dispatcher import CorkDispatcher

from ..server.base_app import BaseAppFactory
from ..configuration.containers import BentoMLContainer

feedback_logger = logging.getLogger("bentoml.feedback")
logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from starlette.routing import BaseRoute
    from starlette.requests import Request
    from starlette.responses import Response

    from ..runner import Runner
    from ..tracing import Tracer


class RunnerAppFactory(BaseAppFactory):
    @inject
    def __init__(
        self,
        runner: "Runner",
        instance_id: t.Optional[int] = None,
        tracer: "Tracer" = Provide[BentoMLContainer.tracer],
    ) -> None:
        self.runner = runner
        self.instance_id = instance_id
        self.tracer = tracer

        from starlette.responses import Response

        TooManyRequests = partial(Response, status_code=427)

        options = self.runner.batch_options
        if options.enabled:
            options = self.runner.batch_options
            self.dispatcher = CorkDispatcher(
                max_latency_in_ms=options.max_latency_ms,
                max_batch_size=options.max_batch_size,
                fallback=TooManyRequests,
            )
        else:
            self.dispatcher = None

    @property
    def name(self) -> str:
        return self.runner.name

    @property
    def on_startup(self) -> t.List[t.Callable[[], None]]:
        on_startup = super().on_startup
        on_startup.insert(0, self.runner._impl.setup)  # type: ignore[reportPrivateUsage]
        return on_startup

    @property
    def on_shutdown(self) -> t.List[t.Callable[[], None]]:
        on_shutdown = super().on_shutdown
        if self.dispatcher is not None:
            on_shutdown.insert(0, self.dispatcher.shutdown)
        return on_shutdown

    @property
    def routes(self) -> t.List["BaseRoute"]:
        """
        Setup routes for Runner server, including:

        /healthz        liveness probe endpoint
        /readyz         Readiness probe endpoint
        /metrics        Prometheus metrics endpoint

        /run
        /run_batch
        """
        from starlette.routing import Route

        routes = super().routes
        routes.append(Route("/run_batch", self.async_run_batch, methods=["POST"]))

        if self.dispatcher is not None:
            _func = self.dispatcher(self._async_cork_run)
            routes.append(Route("/run", _func, methods=["POST"]))
        else:
            routes.append(Route("/run", self.async_run, methods=["POST"]))
        return routes

    async def _async_cork_run(
        self, requests: t.Iterable["Request"]
    ) -> t.List["Response"]:
        from starlette.responses import Response

        assert self._is_ready

        params_list = await asyncio.gather(
            *tuple(multipart_to_payload_params(r) for r in requests)
        )
        params = Params.agg(
            params_list,
            lambda i: AutoContainer.payloads_to_batch(
                i,
                batch_axis=self.runner.batch_options.input_batch_axis,
            ),
        )
        batch_ret = await self.runner.async_run_batch(*params.args, **params.kwargs)
        payloads = AutoContainer.batch_to_payloads(
            batch_ret,
            batch_axis=self.runner.batch_options.input_batch_axis,
        )
        return [
            Response(
                payload.data,
                headers={
                    PAYLOAD_META_HEADER: json.dumps(payload.meta),
                    "Server": f"BentoML-Runner/{self.runner.name}/{self.instance_id}",
                },
            )
            for payload in payloads
        ]

    async def async_run(self, request: "Request") -> "Response":
        from starlette.responses import Response

        assert self._is_ready

        params = await multipart_to_payload_params(request)
        params = params.map(AutoContainer.payload_to_single)
        ret = await self.runner.async_run(*params.args, **params.kwargs)
        payload = AutoContainer.single_to_payload(ret)
        return Response(
            payload.data,
            headers={
                PAYLOAD_META_HEADER: json.dumps(payload.meta),
                "Server": f"BentoML-Runner/{self.runner.name}/{self.instance_id}",
            },
        )

    async def async_run_batch(self, request: "Request") -> "Response":
        from starlette.responses import Response

        assert self._is_ready

        params = await multipart_to_payload_params(request)
        params = params.map(AutoContainer.payload_to_batch)
        ret = await self.runner.async_run_batch(*params.args, **params.kwargs)
        payload = AutoContainer.batch_to_payload(ret)
        return Response(
            payload.data,
            headers={
                PAYLOAD_META_HEADER: json.dumps(payload.meta),
                "Server": f"BentoML-Runner/{self.runner.name}/{self.instance_id}",
            },
        )



import os
import shutil
import typing as t
from typing import TYPE_CHECKING

import numpy as np
from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import Model
from ..models import PTH_EXT
from ..utils.pkg import get_pkg_version
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from ..models import ModelStore

try:
    import easyocr

    assert easyocr.__version__ >= "1.4", BentoMLException(
        "Only easyocr>=1.4 is supported by BentoML"
    )

except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """easyocr is required in order to use module `bentoml.easyocr`, install
        easyocr with `pip install easyocr`. For more information, refers to
        https://www.jaided.ai/easyocr/install/
        """
    )

_easyocr_version = get_pkg_version("easyocr")


@inject
def load(
    tag: str,
    gpu: bool = True,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> easyocr.Reader:
    """
    Load a model from BentoML local modelstore with given tag.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        gpu (`bool`):
            Enable GPU support, default to True
        model_store (`~bentoml._internal.models.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of `catboost.core.CatBoostClassifier` or `catboost.core. CatBoostRegressor`
        from BentoML modelstore.

    Examples::
        import bentoml.catboost
        booster = bentoml.catboost.load(
            "my_model:20201012_DE43A2", model_params=dict(model_type="classifier"))
    """  # noqa

    model = model_store.get(tag)
    if model.info.module != __name__:
        raise BentoMLException(  # pragma: no cover
            f"Model {tag} was saved with"
            f" module {model.info.module},"
            f" failed loading with {__name__}."
        )

    return easyocr.Reader(
        model_storage_directory=model.path,
        download_enabled=False,
        gpu=gpu,
        **model.info.options,
    )


@inject
def save(
    name: str,
    model: easyocr.Reader,
    *,
    lang_list: t.Optional[t.List[str]] = None,
    recog_network: t.Optional[str] = "english_g2",
    detect_model: t.Optional[str] = "craft_mlt_25k",
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`easyocr.Reader`):
            Instance of model to be saved
        lang_list (`List[str]`, `optional`, default to `["en"]`):
            should be same as `easyocr.Reader(lang_list=...)`
        recog_network (`str`, `optional`, default to `english_g2`):
            should be same as `easyocr.Reader(recog_network=...)`
        detect_model (`str`, `optional`, default to `craft_mlt_25k`):
            TODO:
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
        import easyocr

        language_list = ["ch_sim", "en"]
        recog_network = "zh_sim_g2"

        model = easyocr.Reader(
            lang_list=language_list,
            download_enabled=True,
            recog_network=recog_network,
        )

        tag = bentoml.easyocr.save(
            "my_easyocr_model",
            reader,
            lang_list=language_list,
            recog_network=recog_network
        )

        # example tag: my_easyocr_model:20211018_B9EABF

        # load the booster back:
        loaded = bentoml.easyocr.load("my_easyocr_model:latest") # or
        loaded = bentoml.easyocr.load(tag)

    """  # noqa

    context: t.Dict[str, t.Any] = {
        "framework_name": "easyocr",
        "pip_dependencies": [f"easyocr=={_easyocr_version}"],
    }
    if lang_list is None:
        lang_list = ["en"]
    options: t.Dict[str, t.Any] = dict(
        lang_list=lang_list,
        recog_network=recog_network,
    )

    _model = Model.create(
        name,
        module=__name__,
        options=options,
        context=context,
        metadata=metadata,
    )

    src_folder: str = model.model_storage_directory

    detect_filename: str = f"{detect_model}{PTH_EXT}"
    if not os.path.exists(_model.path_of(detect_filename)):
        shutil.copyfile(
            os.path.join(src_folder, detect_filename),
            _model.path_of(detect_filename),
        )

        fname: str = f"{recog_network}{PTH_EXT}"
        shutil.copyfile(os.path.join(src_folder, fname), _model.path_of(fname))

    _model.save(model_store)

    return _model.tag


class _EasyOCRRunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        predict_fn_name: str,
        predict_params: t.Optional[t.Dict[str, t.Any]],
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        super().__init__(str(tag), resource_quota, batch_options)
        self._tag = Tag.from_taglike(tag)
        self._predict_fn_name = predict_fn_name
        self._predict_params = predict_params
        self._model_store = model_store

    @property
    def required_models(self) -> t.List[Tag]:
        return [self._tag]

    @property
    def num_concurrency_per_replica(self) -> int:
        return 1

    @property
    def num_replica(self) -> int:
        if self.resource_quota.on_gpu:
            return len(self.resource_quota.gpus)
        return 1

    # pylint: disable=arguments-differ,attribute-defined-outside-init
    def _setup(self) -> None:  # type: ignore[override]
        self._model = load(self._tag, self.resource_quota.on_gpu, self._model_store)
        self._predict_fn = getattr(self._model, self._predict_fn_name)

    # pylint: disable=arguments-differ
    def _run_batch(  # type: ignore[override]
        self, input_data: np.ndarray
    ) -> "np.ndarray":
        res = self._predict_fn(input_data, **self._predict_params)
        return np.asarray(res, dtype=object)


@inject
def load_runner(
    tag: t.Union[str, Tag],
    predict_fn_name: str = "readtext_batched",
    *,
    predict_params: t.Union[None, t.Dict[str, t.Union[str, t.Any]]] = None,
    resource_quota: t.Union[None, t.Dict[str, t.Any]] = None,
    batch_options: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> _EasyOCRRunner:
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.easyocr.load_runner` implements a Runner class that
    wrap around a EasyOCR Reader model, which optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        predict_fn_name (`str`, default to `readtext_batched`):
            Options for inference functions. Default to `readtext_batched`
        booster_params (`t.Dict[str, t.Union[str, int]]`, default to `None`):
            Parameters for boosters. Refers to https://xgboost.readthedocs.io/en/latest/parameter.html
             for more information
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for `bentoml.xgboost` model

    Examples::
        import xgboost as xgb
        import bentoml.xgboost
        import pandas as pd

        input_data = pd.from_csv("/path/to/csv")
        runner = bentoml.xgboost.load_runner("my_model:20201012_DE43A2")
        runner.run(xgb.DMatrix(input_data))
    """  # noqa
    return _EasyOCRRunner(
        tag=tag,
        predict_fn_name=predict_fn_name,
        predict_params=predict_params,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )



import os
import sys
import typing as t
import asyncio
import logging
from typing import TYPE_CHECKING

from simple_di import inject
from simple_di import Provide

from ...exceptions import BentoMLException
from ..server.base_app import BaseAppFactory
from ..service.service import Service
from ..configuration.containers import BentoMLContainer
from ..configuration.containers import BentoServerContainer
from ..io_descriptors.multipart import Multipart

if TYPE_CHECKING:
    from starlette.routing import BaseRoute
    from starlette.requests import Request
    from starlette.responses import Response
    from starlette.middleware import Middleware
    from starlette.applications import Starlette

    from ..tracing import Tracer
    from ..service.inference_api import InferenceAPI
    from ..server.metrics.prometheus import PrometheusClient


feedback_logger = logging.getLogger("bentoml.feedback")
logger = logging.getLogger(__name__)

DEFAULT_INDEX_HTML = """\
<!DOCTYPE html>
<head>
  <link rel="stylesheet" type="text/css" href="static_content/main.css">
  <link rel="stylesheet" type="text/css" href="static_content/readme.css">
  <link rel="stylesheet" type="text/css" href="static_content/swagger-ui.css">
</head>
<body>
  <div id="tab">
    <button
      class="tabLinks active"
      onclick="openTab(event, 'swagger_ui_container')"
      id="defaultOpen"
    >
      Swagger UI
    </button>
    <button class="tabLinks" onclick="openTab(event, 'markdown_readme')">
      ReadMe
    </button>
  </div>
  <script>
    function openTab(evt, tabName) {{
      // Declare all variables
      var i, tabContent, tabLinks;
      // Get all elements with class="tabContent" and hide them
      tabContent = document.getElementsByClassName("tabContent");
      for (i = 0; i < tabContent.length; i++) {{
        tabContent[i].style.display = "none";
      }}

      // Get all elements with class="tabLinks" and remove the class "active"
      tabLinks = document.getElementsByClassName("tabLinks");
      for (i = 0; i < tabLinks.length; i++) {{
        tabLinks[i].className = tabLinks[i].className.replace(" active", "");
      }}

      // Show the current tab, and add an "active" class to the button that opened the
      // tab
      document.getElementById(tabName).style.display = "block";
      evt.currentTarget.className += " active";
    }}
  </script>
  <div id="markdown_readme" class="tabContent"></div>
  <script src="static_content/marked.min.js"></script>
  <script>
    // TODO: Fix readme escape and index page redesign
    var markdownContent = marked(`{readme}`);
    var element = document.getElementById('markdown_readme');
    element.innerHTML = markdownContent;
  </script>
  <div id="swagger_ui_container" class="tabContent" style="display: block"></div>
  <script src="static_content/swagger-ui-bundle.js"></script>
  <script>
      SwaggerUIBundle({{
          url: '/docs.json',
          dom_id: '#swagger_ui_container'
      }})
  </script>
</body>
"""


def log_exception(request: "Request", exc_info: t.Any) -> None:
    """
    Logs an exception.  This is called by :meth:`handle_exception`
    if debugging is disabled and right before the handler is called.
    The default implementation logs the exception as error on the
    :attr:`logger`.
    """
    logger.error(
        "Exception on %s [%s]", request.url.path, request.method, exc_info=exc_info
    )


class ServiceAppFactory(BaseAppFactory):
    """
    ServiceApp creates a REST API server based on APIs defined with a BentoService
    via BentoService#get_service_apis call. Each InferenceAPI will become one
    endpoint exposed on the REST server, and the RequestHandler defined on each
    InferenceAPI object will be used to handle Request object before feeding the
    request data into a Service API function
    """

    @inject
    def __init__(
        self,
        bento_service: Service,
        enable_metrics: bool = Provide[BentoServerContainer.config.metrics.enabled],
        tracer: "Tracer" = Provide[BentoMLContainer.tracer],
        metrics_client: "PrometheusClient" = Provide[
            BentoServerContainer.metrics_client
        ],
        enable_access_control: bool = Provide[BentoServerContainer.config.cors.enabled],
        access_control_options: t.Dict = Provide[
            BentoServerContainer.access_control_options
        ],
    ) -> None:
        self.bento_service = bento_service

        self.enable_metrics = enable_metrics
        self.tracer = tracer
        self.metrics_client = metrics_client
        self.enable_access_control = enable_access_control
        self.access_control_options = access_control_options

    @property
    def name(self) -> str:
        return self.bento_service.name

    async def index_view_func(
        self, _: "Request"
    ) -> "Response":  # pylint: disable=unused-argument
        """
        The default index view for BentoML API server. This includes the readme
        generated from docstring and swagger UI
        """
        from starlette.responses import Response

        return Response(
            content=DEFAULT_INDEX_HTML.format(readme=self.bento_service.doc),
            status_code=200,
            media_type="text/html",
        )

    @inject
    async def metrics_view_func(
        self,
        _: "Request",
    ) -> "Response":  # pylint: disable=unused-argument
        from starlette.responses import Response

        return Response(
            self.metrics_client.generate_latest(),
            media_type=self.metrics_client.CONTENT_TYPE_LATEST,
        )

    async def docs_view_func(
        self, _: "Request"
    ) -> "Response":  # pylint: disable=unused-argument
        from starlette.responses import JSONResponse

        docs = self.bento_service.openapi_doc()
        return JSONResponse(docs)

    @property
    def routes(self) -> t.List["BaseRoute"]:
        """
        Setup routes for bento model server, including:

        /               Index Page, shows readme docs, metadata, and Swagger UI
        /docs.json      Returns Swagger/OpenAPI definition file in json format
        /healthz        liveness probe endpoint
        /readyz         Readiness probe endpoint
        /metrics        Prometheus metrics endpoint

        And user defined InferenceAPI list into routes, e.g.:
        /classify
        /predict
        """
        from starlette.routing import Mount
        from starlette.routing import Route
        from starlette.staticfiles import StaticFiles

        routes = super().routes

        routes.append(Route(path="/", name="home", endpoint=self.index_view_func))
        routes.append(
            Route(
                path="/docs.json",
                name="docs",
                endpoint=self.docs_view_func,
            )
        )

        if self.enable_metrics:
            routes.append(
                Route(path="/metrics", name="metrics", endpoint=self.metrics_view_func)
            )

        parent_dir_path = os.path.dirname(os.path.realpath(__file__))
        routes.append(
            Mount(
                "/static_content",
                app=StaticFiles(
                    directory=os.path.join(parent_dir_path, "static_content")
                ),
                name="static_content",
            )
        )

        for _, api in self.bento_service._apis.items():
            api_route_endpoint = self._create_api_endpoint(api)
            routes.append(
                Route(
                    path="/{}".format(api.route),
                    name=api.name,
                    endpoint=api_route_endpoint,
                    methods=api.input.HTTP_METHODS,
                )
            )

        return routes

    @property
    def middlewares(self) -> t.List["Middleware"]:
        middlewares = super().middlewares

        from starlette.middleware import Middleware

        for middleware_cls, options in self.bento_service._middlewares:
            middlewares.append(Middleware(middleware_cls, **options))

        if self.enable_access_control:
            assert (
                self.access_control_options.get("allow_origins") is not None
            ), "To enable cors, access_control_allow_origin must be set"

            from starlette.middleware.cors import CORSMiddleware

            middlewares.append(
                Middleware(CORSMiddleware, **self.access_control_options)
            )

        return middlewares

    @property
    def on_startup(self) -> t.List[t.Callable[[], None]]:
        on_startup = super().on_startup
        on_startup.insert(0, self.bento_service._on_asgi_app_startup)
        return on_startup

    @property
    def on_shutdown(self) -> t.List[t.Callable[[], None]]:
        on_shutdown = super().on_shutdown
        on_shutdown.insert(0, self.bento_service._on_asgi_app_shutdown)
        return on_shutdown

    def __call__(self) -> "Starlette":
        app = super().__call__()

        for mount_app, path, name in self.bento_service._mount_apps:
            app.mount(
                app=mount_app, path=path, name=name
            )  # TODO(jiang): mount wsgi app ?

        return app

    @staticmethod
    def _create_api_endpoint(
        api: "InferenceAPI",
    ) -> t.Callable[["Request"], t.Coroutine[t.Any, t.Any, "Response"]]:
        """
        Create api function for flask route, it wraps around user defined API
        callback and adapter class, and adds request logging and instrument metrics
        """
        from starlette.responses import JSONResponse
        from starlette.concurrency import run_in_threadpool

        async def api_func(
            request: "Request",
        ) -> "Response":
            # handle_request may raise 4xx or 5xx exception.
            try:
                input_data = await api.input.from_http_request(request)
                if asyncio.iscoroutinefunction(api.func):
                    if isinstance(api.input, Multipart):
                        output = await api.func(**input_data)
                    else:
                        output = await api.func(input_data)
                else:
                    if isinstance(api.input, Multipart):
                        output = await run_in_threadpool(api.func, **input_data)
                    else:
                        output = await run_in_threadpool(api.func, input_data)
                response = await api.output.to_http_response(output)
            except BentoMLException as e:
                log_exception(request, sys.exc_info())

                if 400 <= e.error_code < 500 and e.error_code not in (401, 403):
                    response = JSONResponse(
                        content="BentoService error handling API request: %s" % str(e),
                        status_code=e.error_code,
                    )
                else:
                    response = JSONResponse("", status_code=e.error_code)
            except Exception:  # pylint: disable=broad-except
                # For all unexpected error, return 500 by default. For example,
                # if users' model raises an error of division by zero.
                log_exception(request, sys.exc_info())

                response = JSONResponse(
                    "An error has occurred in BentoML user code when handling this request, find the error details in server logs",
                    status_code=500,
                )

            return response

        """
        TODO: instrument tracing
        def api_func_with_tracing():
            with self.tracer.span(
                service_name=f"BentoService.{self.bento_service.name}",
                span_name=f"InferenceAPI {api.name} HTTP route",
                request_headers=request.headers,
            ):
                return api_func()

        return api_func_with_tracing
        """
        return api_func



import typing as t
import logging
from io import BytesIO

from starlette.requests import Request
from multipart.multipart import parse_options_header
from starlette.responses import Response

from .base import IODescriptor
from ..types import FileLike
from ...exceptions import BentoMLException

logger = logging.getLogger(__name__)


class File(IODescriptor[FileLike]):
    """
    `File` defines API specification for the inputs/outputs of a Service, where either
     inputs will be converted to or outputs will be converted from file-like objects as
     specified in your API function signature.

    .. Toy implementation of a ViT service::
        # vit_svc.py
        import bentoml
        from bentoml.io import File

        svc = bentoml.Service("vit-object-detection")

        @svc.api(input=File(), output=File())
        def predict(input_pdf):
            return input_pdf

    Users then can then serve this service with `bentoml serve`::
        % bentoml serve ./vit_svc.py:svc --auto-reload

        (Press CTRL+C to quit)
        [INFO] Starting BentoML API server in development mode with auto-reload enabled
        [INFO] Serving BentoML Service "vit-object-detection" defined in "vit_svc.py"
        [INFO] API Server running on http://0.0.0.0:5000

    Users can then send a cURL requests like shown in different terminal session with an
     input PDF files::
        % curl -H "Content-Type: multipart/form-data" -F
          'fileobj=@test.pdf;type=application/pdf' http://0.0.0.0:5000/predict

        %PDF-1.7
                  zed 1/L 1959874/O 282/E 157204/N 12/T 1959312/H [ 500 221]>>
        <</DecodeParms<</Columns 5/Predictor 12>>/Filter/FlateDecode/ID[<5901CB390F31F2B4497ED82C610CB725><7DF08A87AED4DE458FEBB8B3350A46FF>]/Index[280 35]/Info 279 0 R/Length 88/Prev 1959313/Root 281 0 R/Size 315/Type/XRef/W[1 3 1]>>stream
        Warning: Binary output can mess up your terminal. Use "--output -" to tell
        Warning: curl to output it to your terminal anyway, or consider "--output
        Warning: <FILE>" to save to a file.

    Returns:
        IO Descriptor that represents file-like objects.
    """  # noqa: LN001

    def __init__(self, mime_type: t.Optional[str] = None):
        self._mime_type = (
            mime_type if mime_type is not None else "application/octet-stream"
        )

    def openapi_schema_type(self) -> t.Dict[str, str]:
        return {"type": "string", "format": "binary"}

    def openapi_request_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for incoming requests"""
        return {self._mime_type: {"schema": self.openapi_schema_type()}}

    def openapi_responses_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for outcoming responses"""
        return {self._mime_type: {"schema": self.openapi_schema_type()}}

    async def from_http_request(self, request: Request) -> FileLike:
        content_type, _ = parse_options_header(request.headers["content-type"])
        if content_type.decode("utf-8") == "multipart/form-data":
            form = await request.form()
            f = next(iter(form.values()))
            content = await f.read()
            return FileLike(bytes_=content, name=f.filename)
        if content_type.decode("utf-8") == "application/octet-stream":
            body = await request.body()
            return FileLike(bytes_=body)
        raise BentoMLException(
            f"{self.__class__.__name__} should have Content-Type"
            f" b'application/octet-stream' or b'multipart/form-data',"
            f" got {content_type} instead"
        )

    async def to_http_response(self, obj: t.Union[FileLike, bytes]) -> Response:
        if isinstance(obj, bytes):
            obj = FileLike(bytes_=obj)
        return Response(
            t.cast(BytesIO, obj.stream).getvalue(), media_type=self._mime_type
        )



import typing as t

from starlette.requests import Request
from starlette.responses import Response

from .base import IODescriptor

MIME_TYPE = "text/plain"


class Text(IODescriptor[str]):
    """
    `Text` defines API specification for the inputs/outputs of a Service. `Text`
      represents strings for all incoming requests/outcoming responses as specified in
      your API function signature.

    .. Toy implementation of a GPT2 service::
        # gpt2_svc.py
        import bentoml
        from bentoml.io import Text
        import bentoml.transformers

        # If you don't have a gpt2 model previously saved under BentoML modelstore
        # tag = bentoml.transformers.import_from_huggingface_hub('gpt2')
        runner = bentoml.transformers.load_runner('gpt2',tasks='text-generation')

        svc = bentoml.Service("gpt2-generation", runners=[runner])

        @svc.api(input=Text(), output=Text())
        def predict(input_arr):
            res = runner.run_batch(input_arr)
            return res[0]['generated_text']

    Users then can then serve this service with `bentoml serve`::
        % bentoml serve ./gpt2_svc.py:svc --auto-reload

        (Press CTRL+C to quit)
        [INFO] Starting BentoML API server in development mode with auto-reload enabled
        [INFO] Serving BentoML Service "gpt2-generation" defined in "gpt2_svc.py"
        [INFO] API Server running on http://0.0.0.0:5000

    Users can then send a cURL requests like shown in different terminal session::
        % curl -X POST -H "Content-Type: text/plain" --data 'Not for nothing did Orin
         say that people outdoors down here just scuttle in vectors from air
         conditioning to air conditioning.' http://0.0.0.0:5000/predict

        Not for nothing did Orin say that people outdoors down here just scuttle in
         vectors from air conditioning to air conditioning. How do you get to such a
         situation?\n\nWell, you want it to just be one giant monster and that is%

    .. notes::
        `Text` is not designed to take any `args` or `kwargs` during initialization

    Returns:
        IO Descriptor that represents strings type.
    """

    def openapi_schema_type(self) -> t.Dict[str, t.Any]:
        return {"type": "string"}

    def openapi_request_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for incoming requests"""
        return {MIME_TYPE: {"schema": self.openapi_schema_type()}}

    def openapi_responses_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for outcoming responses"""
        return {MIME_TYPE: {"schema": self.openapi_schema_type()}}

    async def from_http_request(self, request: Request) -> str:
        obj = await request.body()
        return str(obj.decode("utf-8"))

    async def to_http_response(self, obj: str) -> Response:
        return Response(obj, media_type=MIME_TYPE)



import typing as t
import logging
from sys import version_info as pyver
from typing import TYPE_CHECKING
from datetime import datetime
from datetime import timezone

import fs
import attr
import yaml
import fs.errors
import fs.mirror
import cloudpickle
from fs.base import FS
from simple_di import inject
from simple_di import Provide

from ..store import Store
from ..store import StoreItem
from ..types import Tag
from ...exceptions import NotFound
from ...exceptions import BentoMLException
from ..configuration import BENTOML_VERSION
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from ..types import PathType
    from ..runner import Runner

logger = logging.getLogger(__name__)

PYTHON_VERSION: str = f"{pyver.major}.{pyver.minor}.{pyver.micro}"
MODEL_YAML_FILENAME = "model.yaml"
CUSTOM_OBJECTS_FILENAME = "custom_objects.pkl"


@attr.define(repr=False)
class Model(StoreItem):
    _tag: Tag
    _fs: FS

    info: "ModelInfo"
    _custom_objects: t.Optional[t.Dict[str, t.Any]] = None

    @property
    def tag(self) -> Tag:
        return self._tag

    @property
    def custom_objects(self) -> t.Dict[str, t.Any]:
        if self._custom_objects is None:
            if self._fs.isfile(CUSTOM_OBJECTS_FILENAME):
                with self._fs.open(CUSTOM_OBJECTS_FILENAME, "rb") as cofile:
                    self._custom_objects = cloudpickle.load(cofile)
            else:
                self._custom_objects = {}

        return self._custom_objects

    def __eq__(self, other: "Model") -> bool:
        return self._tag == other._tag

    def __hash__(self) -> int:
        return hash(self._tag)

    @staticmethod
    def create(
        name: str,
        *,
        module: str = "",
        labels: t.Optional[t.Dict[str, t.Any]] = None,
        options: t.Optional[t.Dict[str, t.Any]] = None,
        custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
        metadata: t.Optional[t.Dict[str, t.Any]] = None,
        context: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> "Model":
        """Create a new Model instance in temporary filesystem used for serializing
        model artifacts and save to model store

        Args:
            name: model name in target model store, model version will be automatically
                generated
            module: import path of module used for saving/loading this model, e.g.
                "bentoml.tensorflow"
            labels:  user-defined labels for managing models, e.g. team=nlp, stage=dev
            options: default options for loading this model, defined by runner
                implementation, e.g. xgboost booster_params
            custom_objects: user-defined additional python objects to be saved
                alongside the model, e.g. a tokenizer instance, preprocessor function,
                model configuration json
            metadata: user-defined metadata for storing model training context
                information or model evaluation metrics, e.g. dataset version,
                training parameters, model scores
            context: Environment context managed by BentoML for loading model,
                e.g. {"framework:" "tensorflow", "framework_version": _tf_version}

        Returns:
            object: Model instance created in temporary filesystem
        """
        tag = Tag(name).make_new_version()
        labels = {} if labels is None else labels
        options = {} if options is None else options
        metadata = {} if metadata is None else metadata
        context = {} if context is None else context
        context["bentoml_version"] = BENTOML_VERSION
        context["python_version"] = PYTHON_VERSION

        model_fs = fs.open_fs(f"temp://bentoml_model_{name}")

        res = Model(
            tag,
            model_fs,
            ModelInfo(
                tag=tag,
                module=module,
                labels=labels,
                options=options,
                metadata=metadata,
                context=context,
            ),
            custom_objects,
        )

        return res

    @inject
    def save(
        self, model_store: "ModelStore" = Provide[BentoMLContainer.model_store]
    ) -> "Model":
        self.flush_info()
        self.flush_custom_objects()

        if not self.validate():
            logger.warning(f"Failed to create Model for {self.tag}, not saving.")
            raise BentoMLException("Failed to save Model because it was invalid")

        with model_store.register(self.tag) as model_path:
            out_fs = fs.open_fs(model_path, create=True, writeable=True)
            fs.mirror.mirror(self._fs, out_fs, copy_if_newer=False)
            self._fs.close()
            self._fs = out_fs

        logger.info(f"Successfully saved {self}")
        return self

    @classmethod
    def from_fs(cls, item_fs: FS) -> "Model":
        try:
            with item_fs.open(MODEL_YAML_FILENAME, "r") as model_yaml:
                info = ModelInfo.from_yaml_file(model_yaml)
        except fs.errors.ResourceNotFound:
            logger.warning(f"Failed to import Model from {item_fs}.")
            raise BentoMLException("Failed to create Model because it was invalid")

        res = cls(info.tag, item_fs, info)
        if not res.validate():
            logger.warning(f"Failed to import Model from {item_fs}.")
            raise BentoMLException("Failed to create Model because it was invalid")

        return res

    @property
    def path(self) -> str:
        return self.path_of("/")

    def path_of(self, item: str) -> str:
        return self._fs.getsyspath(item)

    def flush_info(self):
        with self._fs.open(MODEL_YAML_FILENAME, "w") as model_yaml:
            self.info.dump(model_yaml)

    def flush_custom_objects(self):
        # pickle custom_objects if it is not None and not empty
        if self.custom_objects:
            with self._fs.open(CUSTOM_OBJECTS_FILENAME, "wb") as cofile:
                cloudpickle.dump(self.custom_objects, cofile)

    @property
    def creation_time(self) -> datetime:
        return self.info.creation_time

    def export(self, path: str):
        out_fs = fs.open_fs(path, create=True, writeable=True)
        self.flush_info()
        self.flush_custom_objects()
        fs.mirror.mirror(self._fs, out_fs, copy_if_newer=False)
        out_fs.close()

    def load_runner(self) -> "Runner":
        raise NotImplementedError

    def validate(self):
        return self._fs.isfile(MODEL_YAML_FILENAME)

    def __str__(self):
        return f'Model(tag="{self.tag}", path="{self.path}")'


class ModelStore(Store[Model]):
    def __init__(self, base_path: "t.Union[PathType, FS]"):
        super().__init__(base_path, Model)


@attr.define(repr=False)
class ModelInfo:
    tag: Tag
    module: str
    labels: t.Dict[str, t.Any]
    options: t.Dict[str, t.Any]
    metadata: t.Dict[str, t.Any]
    context: t.Dict[str, t.Any]
    bentoml_version: str = BENTOML_VERSION
    api_version: str = "v1"
    creation_time: datetime = attr.field(factory=lambda: datetime.now(timezone.utc))

    def __attrs_post_init__(self):
        self.validate()

    def dump(self, stream: t.IO[t.Any]):
        return yaml.dump(self, stream, sort_keys=False)

    @classmethod
    def from_yaml_file(cls, stream: t.IO[t.Any]):
        try:
            yaml_content = yaml.safe_load(stream)
        except yaml.YAMLError as exc:  # pragma: no cover - simple error handling
            logger.error(exc)
            raise

        if not isinstance(yaml_content, dict):
            raise BentoMLException(f"malformed {MODEL_YAML_FILENAME}")

        yaml_content["tag"] = Tag(
            yaml_content["name"],  # type: ignore
            yaml_content["version"],  # type: ignore
        )
        del yaml_content["name"]
        del yaml_content["version"]

        try:
            model_info = cls(**yaml_content)  # type: ignore
        except TypeError:  # pragma: no cover - simple error handling
            raise BentoMLException(f"unexpected field in {MODEL_YAML_FILENAME}")
        return model_info

    def validate(self):
        # Validate model.yml file schema, content, bentoml version, etc
        # add tests when implemented
        ...


def copy_model(
    model_tag: t.Union[Tag, str],
    *,
    src_model_store: ModelStore,
    target_model_store: ModelStore,
):
    """copy a model from src model store to target modelstore, and do nothing if the
    model tag already exist in target model store
    """
    try:
        target_model_store.get(model_tag)  # if model tag already found in target
        return
    except NotFound:
        pass

    model = src_model_store.get(model_tag)
    model.save(target_model_store)


def _ModelInfo_dumper(dumper: yaml.Dumper, info: ModelInfo) -> yaml.Node:
    return dumper.represent_dict(
        {
            "name": info.tag.name,
            "version": info.tag.version,
            "bentoml_version": info.bentoml_version,
            "creation_time": info.creation_time,
            "api_version": info.api_version,
            "module": info.module,
            "context": info.context,
            "labels": info.labels,
            "options": info.options,
            "metadata": info.metadata,
        },
    )


yaml.add_representer(ModelInfo, _ModelInfo_dumper)



import json
import typing as t
from typing import TYPE_CHECKING
from json.decoder import JSONDecodeError

from simple_di import inject
from simple_di import Provide

from bentoml._internal.runner.container import Payload

from .runner import RunnerImpl
from ..runner.utils import Params
from ..runner.utils import PAYLOAD_META_HEADER
from ..runner.utils import payload_params_to_multipart
from ..configuration.containers import BentoServerContainer

if TYPE_CHECKING:  # pragma: no cover
    from aiohttp import BaseConnector
    from aiohttp.client import ClientSession


class RemoteRunnerClient(RunnerImpl):
    _conn: t.Optional["BaseConnector"] = None
    _client: t.Optional["ClientSession"] = None

    @inject
    def _get_conn(
        self,
        remote_runner_mapping: t.Dict[str, str] = Provide[
            BentoServerContainer.remote_runner_mapping
        ],
    ) -> "BaseConnector":
        import aiohttp

        uds: str = remote_runner_mapping[self._runner.name]
        if self._conn is None or self._conn.closed:
            self._conn = aiohttp.UnixConnector(path=uds)
        return self._conn

    @inject
    def _get_client(
        self,
        timeout_sec: t.Optional[float] = None,
    ) -> "ClientSession":
        import aiohttp

        if self._client is None or self._client.closed:
            jar = aiohttp.DummyCookieJar()
            if timeout_sec is not None:
                timeout = aiohttp.ClientTimeout(total=timeout_sec)
            else:
                DEFAULT_TIMEOUT = aiohttp.ClientTimeout(total=5 * 60)
                timeout = DEFAULT_TIMEOUT
            self._client = aiohttp.ClientSession(
                connector=self._get_conn(),
                auto_decompress=False,
                cookie_jar=jar,
                connector_owner=False,
                timeout=timeout,
            )
        return self._client

    async def _async_req(self, url: str, *args: t.Any, **kwargs: t.Any) -> t.Any:
        from ..runner.container import AutoContainer

        params = Params(*args, **kwargs).map(AutoContainer.single_to_payload)
        multipart = payload_params_to_multipart(params)
        client = self._get_client()
        async with client.post(url, data=multipart) as resp:
            body = await resp.read()
            meta_header = resp.headers[PAYLOAD_META_HEADER]
        try:
            payload = Payload(data=body, meta=json.loads(meta_header))
        except JSONDecodeError:
            raise ValueError(f"Bento payload decode error: {meta_header}")
        return AutoContainer.payload_to_single(payload)

    async def async_run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        url = "http://127.0.0.1:8000/run"
        return await self._async_req(url, *args, **kwargs)

    async def async_run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        url = "http://127.0.0.1:8000/run_batch"
        return await self._async_req(url, *args, **kwargs)

    def run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        import anyio

        return anyio.run(self.async_run, *args, **kwargs)

    def run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        import anyio

        return anyio.run(self.async_run_batch, *args, **kwargs)



from ._internal.frameworks.spacy import load
from ._internal.frameworks.spacy import save
from ._internal.frameworks.spacy import projects
from ._internal.frameworks.spacy import load_runner
from ._internal.frameworks.spacy import load_project

__all__ = ["projects", "load", "load_project", "load_runner", "save"]



import logging
import multiprocessing
from timeit import default_timer
from typing import TYPE_CHECKING

from simple_di import inject
from simple_di import Provide

from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    Lock = multiprocessing.synchronize.Lock

logger = logging.getLogger(__name__)


class InstrumentMiddleware:
    @inject
    def __init__(
        self,
        app,
        bento_service,
        metrics_client=Provide[BentoMLContainer.metrics_client],
    ):
        self.app = app
        self.bento_service = bento_service

        service_name = self.bento_service.name

        self.metrics_request_duration = metrics_client.Histogram(
            name=service_name + "_request_duration_seconds",
            documentation=service_name + " API HTTP request duration in seconds",
            labelnames=["endpoint", "service_version", "http_response_code"],
        )
        self.metrics_request_total = metrics_client.Counter(
            name=service_name + "_request_total",
            documentation="Total number of HTTP requests",
            labelnames=["endpoint", "service_version", "http_response_code"],
        )
        self.metrics_request_in_progress = metrics_client.Gauge(
            name=service_name + "_request_in_progress",
            documentation="Total number of HTTP requests in progress now",
            labelnames=["endpoint", "service_version"],
            multiprocess_mode="livesum",
        )

    def __call__(self, environ, start_response):
        from flask import Request

        req = Request(environ)
        endpoint = req.path
        start_time = default_timer()

        def start_response_wrapper(status, headers, exc_info=None):
            ret = start_response(status, headers, exc_info)
            status_code = int(status.split()[0])

            # instrument request total count
            self.metrics_request_total.labels(
                endpoint=endpoint,
                service_version=self.bento_service.version,
                http_response_code=status_code,
            ).inc()

            # instrument request duration
            total_time = max(default_timer() - start_time, 0)
            self.metrics_request_duration.labels(
                endpoint=endpoint,
                service_version=self.bento_service.version,
                http_response_code=status_code,
            ).observe(total_time)

            return ret

        with self.metrics_request_in_progress.labels(
            endpoint=endpoint, service_version=self.bento_service.version
        ).track_inprogress():
            return self.app(environ, start_response_wrapper)



from .model import Model
from .model import copy_model
from .model import ModelStore

EXPORTED_STORE_PREFIX = "exported"
SAVE_NAMESPACE = "saved_model"
MODEL_YAML_NAMESPACE = "model_details"

H5_EXT = ".h5"
HDF5_EXT = ".hdf5"
JSON_EXT = ".json"
PKL_EXT = ".pkl"
PTH_EXT = ".pth"
PT_EXT = ".pt"
TXT_EXT = ".txt"
YAML_EXT = ".yaml"
MODEL_EXT = ".model"

__all__ = ["Model", "ModelStore", "copy_model"]



import typing as t

from .runner import Runner
from .runner import RunnerImpl
from .runner import RunnerState
from .runner import SimpleRunner
from ..runner.utils import Params
from ..runner.container import AutoContainer


class LocalRunner(RunnerImpl):
    def setup(self) -> None:
        if self._state != RunnerState.INIT:
            return
        self._state = RunnerState.SETTING
        self._runner._setup()  # type: ignore[reportPrivateUsage]
        self._state = RunnerState.SET

    async def async_run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return self.run(*args, **kwargs)

    async def async_run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return self.run_batch(*args, **kwargs)

    def run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        if self._state is RunnerState.INIT:
            self.setup()
        if isinstance(self._runner, Runner):
            params = Params(*args, **kwargs)
            params = params.map(
                lambda i: AutoContainer.singles_to_batch(
                    [i],
                    batch_axis=self._runner.batch_options.input_batch_axis,
                )
            )
            batch_result = self._runner._run_batch(  # type: ignore[reportPrivateUsage]
                *params.args, **params.kwargs
            )
            single_results = AutoContainer.batch_to_singles(
                batch_result,
                batch_axis=self._runner.batch_options.output_batch_axis,
            )
            return single_results[0]

        if isinstance(self._runner, SimpleRunner):
            return self._runner._run(*args, **kwargs)  # type: ignore[reportPrivateUsage]

    def run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        if self._state is RunnerState.INIT:
            self.setup()
        if isinstance(self._runner, Runner):
            return self._runner._run_batch(*args, **kwargs)  # type: ignore[reportPrivateUsage]
        if isinstance(self._runner, SimpleRunner):
            raise RuntimeError("shall not call run_batch on a" " simple runner")



import os
import sys
import time
import shutil
import typing as t
import logging
import tempfile

from simple_di import inject
from simple_di import Provide

from bentoml import load

from ..configuration import get_debug_mode
from ..configuration.containers import BentoServerContainer

logger = logging.getLogger(__name__)

UVICORN_LOGGING_CONFIG: t.Dict[str, t.Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": "%(message)s",
            "use_colors": False,
            "datefmt": "[%X]",
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "fmt": '%(client_addr)s - "%(request_line)s" %(status_code)s',  # noqa: E501
            "use_colors": False,
            "datefmt": "[%X]",
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "rich.logging.RichHandler",
        },
        "access": {
            "formatter": "access",
            "class": "rich.logging.RichHandler",
        },
    },
    "loggers": {
        "uvicorn": {"handlers": [], "level": "INFO"},
        "uvicorn.error": {"handlers": ["default"], "level": "INFO", "propagate": False},
        "uvicorn.access": {"handlers": ["access"], "level": "INFO", "propagate": False},
    },
}


@inject
def serve_development(
    bento_identifier: str,
    working_dir: str,
    port: int = Provide[BentoServerContainer.config.port],
    host: str = Provide[BentoServerContainer.config.host],
    with_ngrok: bool = False,
    reload: bool = False,
    reload_delay: float = 0.25,
) -> None:
    working_dir = os.path.realpath(os.path.expanduser(working_dir))

    from circus.util import DEFAULT_ENDPOINT_SUB  # type: ignore
    from circus.util import DEFAULT_ENDPOINT_DEALER  # type: ignore
    from circus.arbiter import Arbiter  # type: ignore
    from circus.watcher import Watcher  # type: ignore

    env = dict(os.environ)

    watchers: t.List[Watcher] = []
    if with_ngrok:
        cmd_ngrok = """
import bentoml._internal.server
bentoml._internal.server.start_ngrok_server()"""
        watchers.append(
            Watcher(
                name="ngrok",
                cmd=f"{sys.executable} -c '{cmd_ngrok}'",
                env=env,
                numprocesses=1,
                stop_children=True,
                working_dir=working_dir,
            )
        )

    cmd_api_server = f"""
import bentoml._internal.server
bentoml._internal.server.start_dev_api_server(
    "{bento_identifier}",
    port={port},
    host="{host}",
    working_dir="{working_dir}",
    reload={reload},
    reload_delay={reload_delay},
    instance_id=$(CIRCUS.WID),
)"""
    watchers.append(
        Watcher(
            name="api_server",
            cmd=f"{sys.executable} -c '{cmd_api_server}'",
            env=env,
            numprocesses=1,
            stop_children=True,
            working_dir=working_dir,
        )
    )

    arbiter = Arbiter(
        watchers=watchers,
        endpoint=DEFAULT_ENDPOINT_DEALER,
        pubsub_endpoint=DEFAULT_ENDPOINT_SUB,
    )

    arbiter.start()


def start_ngrok_server() -> None:
    from bentoml._internal.utils.flask_ngrok import start_ngrok

    time.sleep(1)
    start_ngrok(BentoServerContainer.config.port.get())


MAX_AF_UNIX_PATH_LENGTH = 103


@inject
def serve_production(
    bento_identifier: str,
    working_dir: str,
    port: int = Provide[BentoServerContainer.config.port],
    host: str = Provide[BentoServerContainer.config.host],
    backlog: int = Provide[BentoServerContainer.config.backlog],
    app_workers: t.Optional[int] = None,
) -> None:
    working_dir = os.path.realpath(os.path.expanduser(working_dir))

    svc = load(bento_identifier, working_dir=working_dir)
    env = dict(os.environ)

    import json

    from circus.util import DEFAULT_ENDPOINT_SUB  # type: ignore
    from circus.util import DEFAULT_ENDPOINT_DEALER  # type: ignore
    from circus.arbiter import Arbiter  # type: ignore
    from circus.sockets import CircusSocket  # type: ignore
    from circus.watcher import Watcher  # type: ignore

    uds_path = tempfile.mkdtemp()
    watchers: t.List[Watcher] = []
    sockets_map: t.Dict[str, CircusSocket] = {}

    for runner_name, runner in svc.runners.items():
        sockets_path = os.path.join(uds_path, f"{id(runner)}.sock")
        assert len(sockets_path) < MAX_AF_UNIX_PATH_LENGTH
        sockets_map[runner_name] = CircusSocket(
            name=runner_name,
            path=sockets_path,
            umask=0,
        )
        cmd_runner = f"""import bentoml._internal.server
bentoml._internal.server.start_prod_runner_server(
    "{bento_identifier}",
    "{runner_name}",
    working_dir="{working_dir}",
    instance_id=$(CIRCUS.WID),
    fd=$(circus.sockets.{runner_name}),
)"""
        watchers.append(
            Watcher(
                name=f"runner_{runner_name}",
                cmd=f"{sys.executable} -c '{cmd_runner}'",
                env=env,
                numprocesses=runner.num_replica,
                stop_children=True,
                use_sockets=True,
                working_dir=working_dir,
            )
        )
    logger.debug("Runner sockets_map: %s", sockets_map)

    cmd_runner_arg = json.dumps({k: f"{v.path}" for k, v in sockets_map.items()})
    cmd_api_server = f"""
import bentoml._internal.server
bentoml._internal.server.start_prod_api_server(
    "{bento_identifier}",
    port={port},
    host="{host}",
    working_dir="{working_dir}",
    instance_id=$(CIRCUS.WID),
    runner_map={cmd_runner_arg},
    backlog={backlog},
)"""
    watchers.append(
        Watcher(
            name="api_server",
            cmd=f"{sys.executable} -c '{cmd_api_server}'",
            env=env,
            numprocesses=app_workers or 1,
            stop_children=True,
            use_sockets=True,
            working_dir=working_dir,
        )
    )

    arbiter = Arbiter(
        watchers=watchers,
        endpoint=DEFAULT_ENDPOINT_DEALER,
        pubsub_endpoint=DEFAULT_ENDPOINT_SUB,
        sockets=[s for s in sockets_map.values()],
    )

    try:
        arbiter.start()
    finally:

        shutil.rmtree(uds_path)


def start_dev_api_server(
    bento_identifier: str,
    port: int,
    host: str,
    working_dir: t.Optional[str] = None,
    reload: bool = False,
    reload_delay: t.Optional[float] = None,
    instance_id: t.Optional[int] = None,  # pylint: disable=unused-argument
):
    import uvicorn  # type: ignore

    log_level = "debug" if get_debug_mode() else "info"
    svc = load(bento_identifier, working_dir=working_dir)
    uvicorn_options = {
        "host": host,
        "port": port,
        "log_level": log_level,
        "reload": reload,
        "reload_delay": reload_delay,
        "log_config": UVICORN_LOGGING_CONFIG,
    }

    if reload:
        # When reload=True, the app parameter in uvicorn.run(app) must be the import str
        asgi_app_import_str = f"{svc._import_str}.asgi_app"  # type: ignore[reportPrivateUsage]
        # TODO: use svc.build_args.include/exclude as default files to watch
        # TODO: watch changes in model store when "latest" model tag is used
        uvicorn.run(asgi_app_import_str, **uvicorn_options)
    else:
        uvicorn.run(svc.asgi_app, **uvicorn_options)  # type: ignore


def start_prod_api_server(
    bento_identifier: str,
    port: int,
    host: str,
    runner_map: t.Dict[str, str],
    backlog: int,
    working_dir: t.Optional[str] = None,
    instance_id: t.Optional[int] = None,  # pylint: disable=unused-argument
):
    import uvicorn  # type: ignore

    log_level = "info"
    BentoServerContainer.remote_runner_mapping.set(runner_map)
    svc = load(bento_identifier, working_dir=working_dir)
    uvicorn_options = {
        "host": host,
        "port": port,
        "log_level": log_level,
        "backlog": backlog,
        "log_config": UVICORN_LOGGING_CONFIG,
    }

    uvicorn.run(svc.asgi_app, **uvicorn_options)  # type: ignore


def start_prod_runner_server(
    bento_identifier: str,
    name: str,
    fd: int,
    working_dir: t.Optional[str] = None,
    instance_id: t.Optional[int] = None,
):
    import uvicorn  # type: ignore

    from bentoml._internal.server.runner_app import RunnerAppFactory

    svc = load(bento_identifier, working_dir=working_dir)
    runner = svc.runners[name]
    app = RunnerAppFactory(runner, instance_id=instance_id)()

    uvicorn.run(
        app, fd=fd, log_level="info", log_config=UVICORN_LOGGING_CONFIG  # type: ignore
    )



import os
import typing as t
import importlib
import importlib.util
from typing import TYPE_CHECKING
from pathlib import Path

import yaml
from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import Model as BentoModel
from ..utils.pkg import get_pkg_version
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:

    import mlflow.pyfunc
    from mlflow.pyfunc import PyFuncModel

    from ..models import ModelStore


try:
    import mlflow
    from mlflow.models import Model
    from mlflow.models.model import MLMODEL_FILE_NAME
    from mlflow.tracking.artifact_utils import _download_artifact_from_uri
except ImportError:
    raise MissingDependencyException(
        """\
        `mlflow` is required to use with `bentoml.mlflow`.
        Instruction: `pip install -U mlflow`
        """
    )

_mlflow_version = get_pkg_version("mlflow")


def _strike(text: str) -> str:
    return "".join(["\u0336{}".format(c) for c in text])


def _validate_file_exists(fname: str, parent: str) -> t.Tuple[bool, str]:
    for path in Path(parent).iterdir():
        if path.is_dir():
            return _validate_file_exists(fname, str(path))
        if path.name == fname:
            return True, str(path)
    return False, ""


@inject
def load(
    tag: t.Union[str, Tag],
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "PyFuncModel":
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of `mlflow.pyfunc.PyFuncModel` from BentoML modelstore.

    Examples::
    """  # noqa
    model = model_store.get(tag)
    mlflow_folder = model.path_of(model.info.options["mlflow_folder"])
    mlmodel_fpath = Path(mlflow_folder, MLMODEL_FILE_NAME)
    if not mlmodel_fpath.exists():
        raise BentoMLException(f"{MLMODEL_FILE_NAME} cannot be found.")
    flavors = Model.load(mlmodel_fpath).flavors  # pragma: no cover
    module = list(flavors.values())[0]["loader_module"]
    loader_module = importlib.import_module(module)
    return loader_module.load_model(mlflow_folder)  # noqa


def save(*args: str, **kwargs: str) -> None:  # noqa # pylint: disable
    raise EnvironmentError(
        f"""\n
BentoML won't provide a `save` API for MLflow.
If you currently working with `mlflow.<flavor>.save_model`, we kindly suggest you
 to replace `mlflow.<flavor>.save_model` with BentoML's `save` API as we also supports
 the majority of ML frameworks that MLflow supports. An example below shows how you can
 migrate your code for PyTorch from MLflow to BentoML::

    - {_strike('import mlflow.pytorch')}
    + import bentoml.pytorch

    # PyTorch model logics
    ...
    model = EmbeddingBag()
    - {_strike('mlflow.pytorch.save_model(model, "embed_bag")')}
    + bentoml.pytorch.save("embed_bag", model)

If you want to import MLflow models from local directory or a given file path, you can
 utilize `bentoml.mlflow.import_from_uri`::

    import mlflow.pytorch
    + import bentoml.mlflow

    path = "./my_pytorch_model"
    mlflow.pytorch.save_model(model, path)
    + tag = bentoml.mlflow.import_from_uri("mlflow_pytorch_model", path)

If your current workflow with MLflow involve `log_model` as well as importing models from MLflow Registry,
 you can import those directly to BentoML modelstore using `bentoml.mlflow.import_from_uri`. We also accept
 MLflow runs syntax, as well as models registry uri.
An example showing how to integrate your current `log_model` with `mlflow.sklearn` to BentoML::

    import mlflow.sklearn
    + import mlflow
    + import bentoml.mlflow

    # Log sklearn model `sk_learn_rfr` and register as version 1
    ...
    reg_name = "sk-learn-random-forest-reg-model"
    artifact_path = "sklearn_model"
    mlflow.sklearn.log_model(
        sk_model=sk_learn_rfr,
        artifact_path=artifact_path,
        registered_model_name=reg_name
    )

    # refers to https://www.mlflow.org/docs/latest/tracking.html#logging-functions
    + current_run = mlflow.active_run().info.run_id
    + uri = "runs:/%s/%s" % (current_run, artifact_path)
    + tag = bentoml.mlflow.import_from_uri("runs_mlflow_sklearn", uri)

An example showing how to import from MLflow models registry to BentoML modelstore. With this usecase, we
 recommend you to load the model into memory first with `mlflow.<flavor>.load_model` then save the model using
 BentoML `save` API::

    import mlflow.sklearn
    + import bentoml.sklearn

    reg_model_name = "sk-learn-random-forest-reg-model"
    model_uri = "models:/%s/1" % reg_model_name
    + loaded = mlflow.sklearn.load_model(model_uri, *args, **kwargs)
    + tag = bentoml.sklearn.save("my_model", loaded, *args, **kwargs)

    # you can also use `bentoml.mlflow.import_from_uri` to import the model directly after
    #  defining `model_uri`
    + import bentoml.mlflow
    + tag = bentoml.mlflow.import_from_uri("my_model", model_uri)
    """  # noqa
    )


@inject
def import_from_uri(
    name: str,
    uri: str,
    *,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    context: t.Dict[str, t.Any] = {
        "framework_name": "mlflow",
        "pip_dependencies": [f"mlflow=={_mlflow_version}"],
    }

    _model = BentoModel.create(
        name,
        module=__name__,
        options=None,
        context=context,
        metadata=metadata,
    )

    _model.info.options = {"uri": uri}
    mlflow_obj_path = _download_artifact_from_uri(uri, output_path=_model.path)
    _model.info.options["mlflow_folder"] = os.path.relpath(mlflow_obj_path, _model.path)
    exists, _ = _validate_file_exists("MLproject", mlflow_obj_path)
    if exists:
        raise BentoMLException("BentoML doesn't accept MLflow Projects.")
    exists, fpath = _validate_file_exists("MLmodel", mlflow_obj_path)
    if not exists:
        raise BentoMLException("Downloaded path is not a valid MLflow Model.")
    with Path(fpath).open("r") as mf:
        conf = yaml.safe_load(mf.read())
        flavor = conf["flavors"]["python_function"]
        _model.info.options["flavor"] = flavor["loader_module"]

    _model.save(model_store)

    return _model.tag


class _PyFuncRunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        super().__init__(str(tag), resource_quota, batch_options)
        self._model_store = model_store
        self._model_info = self._model_store.get(tag)

    @property
    def required_models(self) -> t.List[Tag]:
        return [self._model_info.tag]

    @property
    def num_concurrency_per_replica(self) -> int:
        return int(round(self.resource_quota.cpu))

    @property
    def num_replica(self) -> int:
        return 1

    # pylint: disable=arguments-differ,attribute-defined-outside-init
    def _setup(self) -> None:  # type: ignore[override]
        path = self._model_info.info.options["mlflow_folder"]
        artifact_path = self._model_info.path_of(path)
        self._model = mlflow.pyfunc.load_model(artifact_path, suppress_warnings=False)

    # pylint: disable=arguments-differ
    def _run_batch(self, input_data: t.Any) -> t.Any:  # type: ignore[override]
        return self._model.predict(input_data)


@inject
def load_runner(
    tag: t.Union[str, Tag],
    resource_quota: t.Optional[t.Dict[str, t.Any]] = None,
    batch_options: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "_PyFuncRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.mlflow.load_runner` implements a Runner class that
    has to options to wrap around a PyFuncModel, or infer from given MLflow flavor to
    load BentoML internal Runner implementation, which optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances loaded from `bentoml.mlflow`. This can be either a `mlflow.pyfunc.PyFuncModel`
        Runner or BentoML runner instance that maps to MLflow flavors

    Examples::
    """  # noqa
    return _PyFuncRunner(
        tag,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )



"""
User facing python APIs for managing local bentos and build new bentos
"""
import os
import typing as t
import logging
import subprocess
from typing import TYPE_CHECKING

import fs
from simple_di import inject
from simple_di import Provide

from .exceptions import InvalidArgument
from ._internal.bento import Bento
from ._internal.types import Tag
from ._internal.utils import resolve_user_filepath
from ._internal.bento.build_config import BentoBuildConfig
from ._internal.configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from ._internal.bento import BentoStore
    from ._internal.models import ModelStore

logger = logging.getLogger(__name__)


@inject
def list(  # pylint: disable=redefined-builtin
    tag: t.Optional[t.Union[Tag, str]] = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
) -> "t.List[Bento]":
    return _bento_store.list(tag)


@inject
def get(
    tag: t.Union[Tag, str],
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
) -> Bento:
    return _bento_store.get(tag)


@inject
def delete(
    tag: t.Union[Tag, str],
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
):
    _bento_store.delete(tag)


@inject
def import_bento(
    path: str,
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
) -> Bento:
    return Bento.from_fs(fs.open_fs(path))


@inject
def export_bento(
    tag: t.Union[Tag, str],
    path: str,
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
):
    bento = get(tag, _bento_store=_bento_store)
    bento.export(path)


@inject
def push(
    tag: t.Union[Tag, str],
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
):
    raise NotImplementedError


@inject
def pull(
    tag: t.Union[Tag, str],
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
):
    raise NotImplementedError


@inject
def build(
    service: str,
    *,
    labels: t.Optional[t.Dict[str, str]] = None,
    description: t.Optional[str] = None,
    include: t.Optional[t.List[str]] = None,
    exclude: t.Optional[t.List[str]] = None,
    additional_models: t.Optional[t.List[t.Union[str, Tag]]] = None,
    docker: t.Optional[t.Dict[str, t.Any]] = None,
    python: t.Optional[t.Dict[str, t.Any]] = None,
    conda: t.Optional[t.Dict[str, t.Any]] = None,
    version: t.Optional[str] = None,
    build_ctx: t.Optional[str] = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "Bento":
    """
    User-facing API for building a Bento, the available build options are symmetrical to
    the content of a valid bentofile.yaml file, for building Bento from CLI.

    This API will not respect bentofile.yaml file in current environment, build options
    can only be provided via function call parameters.

    Args:
        service: import str for finding the bentoml.Service instance build target
        labels: optional immutable labels for carrying contextual info
        description: optional description string in markdown format
        include: list of file paths and patterns specifying files to include in Bento,
            default is all files under build_ctx, beside the ones excluded from the
            exclude parameter or a .bentoignore file for a given directory
        exclude: list of file paths and patterns to exclude from the final Bento archive
        additional_models: list of model tags to pack in Bento, in addition to the
            models that are required by service's runners. These models must be
            found in the given _model_store
        docker: dictionary for configuring Bento's containerization process, see details
            in `._internal.bento.build_config.DockerOptions`
        python: dictionary for configuring Bento's python dependencies, see details in
            `._internal.bento.build_config.PythonOptions`
        conda: dictionary for configuring Bento's conda dependencies, see details in
            `._internal.bento.build_config.CondaOptions`
        version: Override the default auto generated version str
        build_ctx: Build context directory, when used as
        _bento_store: save Bento created to this BentoStore
        _model_store: pull Models required from this ModelStore

    Returns:
        Bento: a Bento instance representing the materialized Bento saved in BentoStore

    """  # noqa: LN001
    build_config = BentoBuildConfig(
        service=service,
        description=description,
        labels=labels,
        include=include,
        exclude=exclude,
        additional_models=additional_models,
        docker=docker,
        python=python,
        conda=conda,
    )

    bento = Bento.create(
        build_config=build_config,
        version=version,
        build_ctx=build_ctx,
        model_store=_model_store,
    ).save(_bento_store)
    logger.info("Bento build success, %s created", bento)
    return bento


@inject
def build_bentofile(
    bentofile: str = "bentofile.yaml",
    *,
    version: t.Optional[str] = None,
    build_ctx: t.Optional[str] = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "Bento":
    """
    Build a Bento base on options specified in a bentofile.yaml file.

    By default, this function will look for a `bentofile.yaml` file in current working
    directory.

    Args:
        bentofile: The file path to build config yaml file
        version: Override the default auto generated version str
        build_ctx: Build context directory, when used as
        _bento_store: save Bento created to this BentoStore
        _model_store: pull Models required from this ModelStore
    """
    try:
        bentofile = resolve_user_filepath(bentofile, build_ctx)
    except FileNotFoundError:
        raise InvalidArgument(f'bentofile "{bentofile}" not found')

    with open(bentofile, "r", encoding="utf-8") as f:
        build_config = BentoBuildConfig.from_yaml(f)

    bento = Bento.create(
        build_config=build_config,
        version=version,
        build_ctx=build_ctx,
        model_store=_model_store,
    ).save(_bento_store)
    logger.info("Bento build success, %s created", bento)
    return bento


@inject
def containerize(
    tag: t.Union[Tag, str],
    docker_image_tag: t.Optional[str] = None,
    *,
    build_args: t.Optional[t.Dict[str, str]] = None,
    labels: t.Optional[t.Dict[str, str]] = None,
    no_cache: bool = False,
    platform: t.Optional[str] = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
):
    bento = _bento_store.get(tag)
    if docker_image_tag is None:
        docker_image_tag = str(bento.tag)

    docker_build_cmd = ["docker", "build", "."]
    dockerfile_path = os.path.join("env", "docker", "Dockerfile")
    docker_build_cmd += ["-f", dockerfile_path]
    docker_build_cmd += ["-t", docker_image_tag]

    if no_cache:
        docker_build_cmd.append("--no-cache")

    if build_args:
        for key, value in build_args.items():
            docker_build_cmd += ["--build-arg", f"{key}={value}"]

    if labels:
        for key, value in labels.items():
            docker_build_cmd += ["--label", f"{key}={value}"]

    if platform:
        docker_build_cmd += ["--platform", platform]

    env = os.environ.copy()
    env["DOCKER_SCAN_SUGGEST"] = "false"
    logger.info(f"Building docker image for {bento}...")
    subprocess.check_output(docker_build_cmd, cwd=bento.path, env=env)
    logger.info(f'Successfully built docker image "{docker_image_tag}"')


__all__ = [
    "list",
    "get",
    "delete",
    "import_bento",
    "export_bento",
    "push",
    "pull",
    "build",
    "build_bentofile",
    "containerize",
]



import typing as t
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

from starlette.requests import Request
from starlette.responses import Response

from ..types import FileLike

if TYPE_CHECKING:  # pragma: no cover
    import numpy as np  # noqa: F401
    import pandas as pd  # noqa: F401
    import pydantic  # noqa: F401
    import PIL.Image  # noqa: F401


JSONType = t.Union[str, t.Dict[str, t.Any], "pydantic.BaseModel"]


# NOTES: we will keep type in quotation to avoid backward compatibility
#  with numpy < 1.20, since we will use the latest stubs from the main branch of numpy.
#  that enable a new way to type hint an ndarray.
ImageType = t.Union["PIL.Image.Image", "np.ndarray[t.Any, np.dtype[t.Any]]"]

IOType = t.Union[
    str,
    JSONType,
    FileLike,
    ImageType,
    "np.ndarray[t.Any, np.dtype[t.Any]]",
    "pd.DataFrame",
    "pd.Series[t.Any]",
]


IOPyObj = t.TypeVar("IOPyObj")


_T = t.TypeVar("_T")


class IODescriptor(ABC, t.Generic[IOPyObj]):
    """
    IODescriptor describes the input/output data format of an InferenceAPI defined
     in a bentoml.Service
    """

    HTTP_METHODS = ["POST"]
    _init_str: str = ""

    def __new__(cls: t.Type[_T], *args: t.Any, **kwargs: t.Any) -> _T:
        self = super().__new__(cls)
        arg_strs = tuple(repr(i) for i in args) + tuple(
            f"{k}={repr(v)}" for k, v in kwargs.items()
        )
        setattr(self, "_init_str", f"{cls.__name__}({', '.join(arg_strs)})")

        return self

    def __repr__(self) -> str:
        return self._init_str

    @abstractmethod
    def openapi_schema_type(self) -> t.Dict[str, str]:
        ...

    @abstractmethod
    def openapi_request_schema(self) -> t.Dict[str, t.Any]:
        ...

    @abstractmethod
    def openapi_responses_schema(self) -> t.Dict[str, t.Any]:
        ...

    @abstractmethod
    async def from_http_request(self, request: Request) -> IOPyObj:
        ...

    @abstractmethod
    async def to_http_response(self, obj: IOPyObj) -> Response:
        ...

    # TODO: gRPC support
    # @abstractmethod
    # def generate_protobuf(self): ...

    # @abstractmethod
    # async def from_grpc_request(self, request: GRPCRequest) -> IOType: ...

    # @abstractmethod
    # async def to_grpc_response(self, obj: IOType) -> GRPCResponse: ...



import os
import re
import math
import ctypes
import typing as t
import logging
import itertools
from typing import TYPE_CHECKING
from functools import lru_cache

from simple_di.providers import SingletonFactory

from ...exceptions import BentoMLException

logger = logging.getLogger(__name__)

# Some constants taken from cuda.h


if TYPE_CHECKING:
    from ctypes import c_int
    from ctypes import c_char_p
    from ctypes import c_size_t
    from ctypes import c_void_p

    CDataType = t.Union[c_int, c_void_p, c_size_t, c_char_p]

    from aiohttp import MultipartWriter
    from starlette.requests import Request

    from ..runner.container import Payload

    class PlcType(t.TypedDict):
        err: c_char_p
        device: c_int
        num_gpus: c_int
        context: c_void_p
        free_mem: c_size_t
        total_mem: c_size_t


T = t.TypeVar("T")
To = t.TypeVar("To")


CUDA_SUCCESS = 0


class Params(t.Generic[T]):
    def __init__(self, *args: T, **kwargs: T):
        self.args: t.Tuple[T, ...] = args
        self.kwargs: t.Dict[str, T] = kwargs

    def map(self, function: t.Callable[[T], To]) -> "Params[To]":
        args = tuple(function(a) for a in self.args)
        kwargs = {k: function(v) for k, v in self.kwargs.items()}
        return Params[To](*args, **kwargs)

    def imap(
        self, function: t.Callable[[T], t.Iterable[To]]
    ) -> "t.Iterator[Params[To]]":
        args_iter = tuple(iter(function(a)) for a in self.args)
        kwargs_iter = {k: iter(function(v)) for k, v in self.kwargs.items()}

        try:
            while True:
                args = tuple(next(a) for a in args_iter)
                kwargs = {k: next(v) for k, v in kwargs_iter.items()}
                yield Params[To](*args, **kwargs)
        except StopIteration:
            pass

    def items(self) -> t.Iterator[t.Tuple[t.Union[int, str], T]]:
        return itertools.chain(enumerate(self.args), self.kwargs.items())

    @classmethod
    def agg(
        cls,
        params_list: t.Sequence["Params[T]"],
        agg_func: t.Callable[[t.Sequence[T]], To] = lambda i: i,
    ) -> "Params[To]":
        if not params_list:
            return t.cast(Params[To], [])

        args: t.List[To] = []
        kwargs: t.Dict[str, To] = {}

        for j, _ in enumerate(params_list[0].args):
            arg: t.List[T] = []
            for params in params_list:
                arg.append(params.args[j])
            args.append(agg_func(arg))
        for k in params_list[0].kwargs:
            kwarg: t.List[T] = []
            for params in params_list:
                kwarg.append(params.kwargs[k])
            kwargs[k] = agg_func(kwarg)
        return Params(*tuple(args), **kwargs)

    @property
    def sample(self) -> T:
        if self.args:
            return self.args[0]
        return next(iter(self.kwargs.values()))


PAYLOAD_META_HEADER = "Bento-Payload-Meta"


def payload_params_to_multipart(params: Params["Payload"]) -> "MultipartWriter":
    import json

    from multidict import CIMultiDict
    from aiohttp.multipart import MultipartWriter

    multipart = MultipartWriter(subtype="form-data")
    for key, payload in params.items():
        multipart.append(
            payload.data,
            headers=CIMultiDict(
                (
                    (PAYLOAD_META_HEADER, json.dumps(payload.meta)),
                    ("Content-Disposition", f'form-data; name="{key}"'),
                )
            ),
        )
    return multipart


async def multipart_to_payload_params(request: "Request") -> Params["Payload"]:
    import json

    from bentoml._internal.runner.container import Payload
    from bentoml._internal.utils.formparser import populate_multipart_requests

    parts = await populate_multipart_requests(request)
    max_arg_index = -1
    kwargs: t.Dict[str, Payload] = {}
    args_map: t.Dict[int, Payload] = {}
    for field_name, req in parts.items():
        payload = Payload(
            data=await req.body(),
            meta=json.loads(req.headers[PAYLOAD_META_HEADER]),
        )
        if field_name.isdigit():
            arg_index = int(field_name)
            args_map[arg_index] = payload
            max_arg_index = max(max_arg_index, arg_index)
        else:
            kwargs[field_name] = payload
    args = tuple(args_map[i] for i in range(max_arg_index + 1))
    return Params(*args, **kwargs)


class TypeRef:
    def __init__(self, module: str, qualname: str):
        self.module = module
        self.qualname = qualname

    @classmethod
    def from_instance(cls, instance: object) -> "TypeRef":
        klass = type(instance)
        return cls.from_type(klass)

    @classmethod
    def from_type(cls, klass: t.Union["TypeRef", type]) -> "TypeRef":
        if isinstance(klass, type):
            return cls(klass.__module__, klass.__qualname__)
        return klass

    def evaluate(self) -> type:
        import importlib

        m = importlib.import_module(self.module)
        ref = t.ForwardRef(f"m.{self.qualname}")
        localns = {"m": m}

        if hasattr(t, "_eval_type"):  # python3.7, 3.8 & 3.9
            _eval_type = getattr(t, "_eval_type")
            return t.cast(type, _eval_type(ref, globals(), localns))

        raise SystemError("unsupported Python version")

    def __eq__(self, o: object) -> bool:
        """
        TypeRef("numpy", "ndarray") == np.ndarray
        TypeRef("numpy", "ndarray") == TypeRef.from_instance(np.random.randint([2, 3]))
        """
        if isinstance(o, type):
            o = self.from_type(o)

        if isinstance(o, TypeRef):
            return self.module == o.module and self.qualname == o.qualname

        return False

    def __hash__(self) -> int:
        return hash(f"{self.module}.{self.qualname}")

    def __repr__(self) -> str:
        return f'TypeRef("{self.module}", "{self.qualname}")'


def cpu_converter(cpu: t.Union[int, float, str]) -> float:
    if isinstance(cpu, (int, float)):
        return float(cpu)

    if isinstance(cpu, str):
        milli_match = re.match("([0-9]+)m", cpu)
        if milli_match:
            return int(milli_match[1]) / 1000.0

    raise ValueError(f"Invalid CPU resource limit '{cpu}'")


def mem_converter(mem: t.Union[int, str]) -> int:
    if isinstance(mem, int):
        return mem

    if isinstance(mem, str):
        unit_match = re.match("([0-9]+)([A-Za-z]{1,2})", mem)
        mem_multipliers = {
            "k": 1000,
            "M": 1000 ** 2,
            "G": 1000 ** 3,
            "T": 1000 ** 4,
            "P": 1000 ** 5,
            "E": 1000 ** 6,
            "Ki": 1024,
            "Mi": 1024 ** 2,
            "Gi": 1024 ** 3,
            "Ti": 1024 ** 4,
            "Pi": 1024 ** 5,
            "Ei": 1024 ** 6,
        }
        if unit_match:
            base = int(unit_match[1])
            unit = unit_match[2]
            if unit in mem_multipliers:
                return base * mem_multipliers[unit]

    raise ValueError(f"Invalid MEM resource limit '{mem}'")


@lru_cache(maxsize=1)
def query_cgroup_cpu_count() -> float:
    # Query active cpu processor count using cgroup v1 API, based on OpenJDK
    # implementation for `active_processor_count` using cgroup v1:
    # https://github.com/openjdk/jdk/blob/master/src/hotspot/os/linux/cgroupSubsystem_linux.cpp
    # For cgroup v2, see:
    # https://github.com/openjdk/jdk/blob/master/src/hotspot/os/linux/cgroupV2Subsystem_linux.cpp
    def _read_integer_file(filename: str) -> int:
        with open(filename, "r", encoding="utf-8") as f:
            return int(f.read().rstrip())

    cgroup_root = "/sys/fs/cgroup/"
    cfs_quota_us_file = os.path.join(cgroup_root, "cpu", "cpu.cfs_quota_us")
    cfs_period_us_file = os.path.join(cgroup_root, "cpu", "cpu.cfs_period_us")
    shares_file = os.path.join(cgroup_root, "cpu", "cpu.shares")

    quota = shares = period = -1
    if os.path.isfile(cfs_quota_us_file):
        quota = _read_integer_file(cfs_quota_us_file)

    if os.path.isfile(shares_file):
        shares = _read_integer_file(shares_file)
        if shares == 1024:
            shares = -1

    if os.path.isfile(cfs_period_us_file):
        period = _read_integer_file(cfs_period_us_file)

    os_cpu_count = float(os.cpu_count() or 1)

    limit_count = math.inf
    quota_count = 0.0
    share_count = 0.0

    if quota > -1 and period > 0:
        quota_count = float(quota) / float(period)
    if shares > -1:
        share_count = float(shares) / float(1024)

    if quota_count != 0 and share_count != 0:
        limit_count = min(quota_count, share_count)
    if quota_count != 0:
        limit_count = quota_count
    if share_count != 0:
        limit_count = share_count

    return float(min(limit_count, os_cpu_count))


@SingletonFactory
def _cuda_lib() -> "ctypes.CDLL":
    libs = ("libcuda.so", "cuda.dll")
    for lib in libs:
        try:
            return ctypes.CDLL(lib)
        except OSError:
            continue
    raise OSError(f"could not load any of: {' '.join(libs)}")


@SingletonFactory
def _init_var() -> t.Tuple["ctypes.CDLL", "PlcType"]:
    # https://docs.nvidia.com/cuda/cuda-driver-api/group__CUDA__DEVICE.html
    # TODO: add threads_per_core, cores, Compute Capability
    err = ctypes.c_char_p()
    plc: PlcType = {
        "err": err,
        "device": ctypes.c_int(),
        "num_gpus": ctypes.c_int(),
        "context": ctypes.c_void_p(),
        "free_mem": ctypes.c_size_t(),
        "total_mem": ctypes.c_size_t(),
    }

    try:
        _drv = _cuda_lib.get()
        res = _drv.cuInit(0)
        if res != CUDA_SUCCESS:
            _drv.cuGetErrorString(res, ctypes.byref(err))
            logger.error(f"cuInit failed with error code {res}: {str(err.value)}")
        return _drv, plc
    except OSError as e:
        raise BentoMLException(
            f"{e}\nMake sure to have CUDA "
            f"installed you are intending "
            f"to use GPUs with BentoML."
        )


def gpu_converter(gpus: t.Optional[t.Union[int, str, t.List[str]]]) -> t.List[str]:
    if gpus is not None:
        drv, plc = _init_var.get()

        res = drv.cuDeviceGetCount(ctypes.byref(plc["num_gpus"]))
        if res != CUDA_SUCCESS:
            drv.cuGetErrorString(res, ctypes.byref(plc["err"]))
            logger.error(
                f"cuDeviceGetCount failed with error code {res}: {str(plc['err'].value)}"
            )

        def _validate_dev(dev_id: t.Union[int, str]) -> bool:
            _res = drv.cuDeviceGet(ctypes.byref(plc["device"]), int(dev_id))
            if _res != CUDA_SUCCESS:
                drv.cuGetErrorString(_res, ctypes.byref(plc["err"]))
                logger.warning(
                    "cuDeviceGet failed "
                    f"with error code {_res}: {str(plc['err'].value)}"
                )
                return False
            return True

        if isinstance(gpus, (int, str)):
            if gpus == "all":
                return [str(dev) for dev in range(plc["num_gpus"].value)]
            else:
                if _validate_dev(gpus):
                    return [str(gpus)]
                raise BentoMLException(
                    f"Unknown GPU devices. Available devices: {plc['num_gpus'].value}"
                )
        else:
            return list(
                itertools.chain.from_iterable([gpu_converter(gpu) for gpu in gpus])
            )
    return list()


def get_gpu_memory(dev: int) -> t.Tuple[int, int]:
    """Return Total Memory and Free Memory in given GPU device. in MiB"""
    drv, plc = _init_var.get()

    res = drv.cuDeviceGet(ctypes.byref(plc["device"]), dev)
    if res != CUDA_SUCCESS:
        drv.cuGetErrorString(res, ctypes.byref(plc["err"]))
        logger.error(
            "cuDeviceGet failed " f"with error code {res}: {str(plc['err'].value)}"
        )
    try:
        res = drv.cuCtxCreate_v2(ctypes.byref(plc["context"]), 0, plc["device"])
    except AttributeError:
        res = drv.cuCtxCreate(ctypes.byref(plc["context"]), 0, plc["device"])
    if res != CUDA_SUCCESS:
        drv.cuGetErrorString(res, ctypes.byref(plc["err"]))
        logger.error(
            f"cuCtxCreate failed with error code {res}: {str(plc['err'].value)}"
        )

    try:
        res = drv.cuMemGetInfo_v2(
            ctypes.byref(plc["free_mem"]), ctypes.byref(plc["total_mem"])
        )
    except AttributeError:
        res = drv.cuMemGetInfo(
            ctypes.byref(plc["free_mem"]), ctypes.byref(plc["total_mem"])
        )
    if res != CUDA_SUCCESS:
        drv.cuGetErrorString(res, ctypes.byref(plc["err"]))
        logger.error(
            f"cuMemGetInfo failed with error code {res}: " f"{str(plc['err'].value)}"
        )
    _total_mem = plc["total_mem"].value
    _free_mem = plc["free_mem"].value
    logger.debug(f"Total Memory: {_total_mem} MiB\nFree Memory: {_free_mem} MiB")
    drv.cuCtxDetach(plc["context"])
    return _total_mem, _free_mem



import re
import typing as t
from typing import Optional

import yaml

from ...exceptions import InvalidArgument
from ..io_descriptors import IODescriptor

RESERVED_API_NAMES = [
    "index",
    "swagger",
    "docs",
    "metrics",
    "healthz",
    "livez",
    "readyz",
]


class InferenceAPI:
    def __init__(
        self,
        user_defined_callback: t.Callable[..., t.Any],
        input_descriptor: IODescriptor[t.Any],
        output_descriptor: IODescriptor[t.Any],
        name: Optional[str],
        doc: Optional[str] = None,
        route: Optional[str] = None,
    ):
        # Use user_defined_callback function variable if name not specified
        name = user_defined_callback.__name__ if name is None else name
        # Use user_defined_callback function docstring `__doc__` if doc not specified
        doc = user_defined_callback.__doc__ if doc is None else doc
        # Use API name as route if route not specified
        route = name if route is None else route

        InferenceAPI._validate_name(name)
        InferenceAPI._validate_route(route)

        self.name = name
        self.func = user_defined_callback
        self.input = input_descriptor
        self.output = output_descriptor
        self.doc = doc
        self.route = route

    def __str__(self):
        return f"{self.__class__.__name__}({str(self.input)} → {str(self.output)})"

    @staticmethod
    def _validate_name(api_name: str):
        if not api_name.isidentifier():
            raise InvalidArgument(
                "Invalid API name: '{}', a valid identifier may only contain letters,"
                " numbers, underscores and not starting with a number.".format(api_name)
            )

        if api_name in RESERVED_API_NAMES:
            raise InvalidArgument(
                "Reserved API name: '{}' is reserved for infra endpoints".format(
                    api_name
                )
            )

    @staticmethod
    def _validate_route(route: str):
        if re.findall(
            r"[?#]+|^(//)|^:", route
        ):  # contains '?' or '#' OR  start with '//' OR start with ':'
            # https://tools.ietf.org/html/rfc3986#page-22
            raise InvalidArgument(
                "The path {} contains illegal url characters".format(route)
            )
        if route in RESERVED_API_NAMES:
            raise InvalidArgument(
                "Reserved API route: '{}' is reserved for infra endpoints".format(route)
            )


def _InferenceAPI_dumper(dumper: yaml.Dumper, api: InferenceAPI) -> yaml.Node:
    return dumper.represent_dict(
        {
            "route": api.route,
            "doc": api.doc,
            "input": api.input.__class__.__name__,
            "output": api.output.__class__.__name__,
        }
    )


yaml.add_representer(InferenceAPI, _InferenceAPI_dumper)
